from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from rex.modules.embeddings.static_embedding import StaticEmbedding
from rex.modules.pcnn import PiecewiseCNN


class PCNNOne(nn.Module):
    """
    PCNN + ONE for bag-level multi-class relation classification

    References:
        - Zeng D et al. Distant supervision for relation extraction via piecewise convolutional neural networks. EMNLP. 2015.
    """

    def __init__(
        self,
        vocab,
        num_classes,
        dim_token_emb,
        pos_emb_capacity,
        dim_pos,
        num_filters,
        kernel_size,
        dropout: Optional[float] = 0.5,
    ):
        super().__init__()

        self.token_embedding = StaticEmbedding(
            vocab.size, dim_token_emb, dropout=dropout
        )
        self.pos1_embedding = nn.Embedding(
            num_embeddings=pos_emb_capacity, embedding_dim=dim_pos
        )
        self.pos2_embedding = nn.Embedding(
            num_embeddings=pos_emb_capacity, embedding_dim=dim_pos
        )
        self.pcnn = PiecewiseCNN(
            dim_token_emb + 2 * dim_pos, num_filters, kernel_size, dropout=dropout
        )
        self.dense = nn.Linear(
            in_features=num_filters * 3, out_features=num_classes, bias=True
        )

    def forward(self, token_ids, head_pos, tail_pos, mask, scopes, labels=None):
        token_embedding = self.token_embedding(token_ids)
        pos1_embedding = self.pos1_embedding(head_pos)
        pos2_embedding = self.pos2_embedding(tail_pos)
        x = torch.cat([token_embedding, pos1_embedding, pos2_embedding], dim=-1)
        x = self.pcnn(x, mask)

        if labels is not None:
            bag_rep_for_loss = []
        bag_rep_for_pred = []
        for ind, scope in enumerate(scopes):
            sent_rep = x[scope[0] : scope[1], :]
            out = self.dense(sent_rep)
            probs = torch.softmax(out, dim=-1)

            # for training loss calculation
            if labels is not None:
                _, j = torch.max(probs[:, labels[ind]], dim=-1)
                bag_rep_for_loss.append(out[j])

            # for prediction
            row_prob, row_idx = torch.max(probs, dim=-1)
            if row_idx.sum() > 0:
                mask = row_idx.view(-1, 1).expand(-1, probs.shape[-1])
                probs = probs.masked_fill_(mask.eq(0), float("-inf"))
                row_prob, _ = torch.max(probs[:, 1:], dim=-1)
                _, row_idx = torch.max(row_prob, dim=0)
            else:
                _, row_idx = torch.max(row_prob, dim=-1)
            bag_rep_for_pred.append(out[row_idx])

        result = {}
        if labels is not None:
            bag_rep_for_loss = torch.stack(bag_rep_for_loss, dim=0)
            result.update({"loss": F.cross_entropy(bag_rep_for_loss, labels)})
        bag_rep_for_pred = torch.stack(bag_rep_for_pred, dim=0)
        result.update({"pred": bag_rep_for_pred.argmax(dim=-1)})

        return result


class PCNNAtt(nn.Module):
    """
    PCNN + Selective Attention for bag-level multi-class relation classification

    References:
        - Lin Y et al. Neural relation extraction with selective attention over instances. ACL. 2016.
    """

    def __init__(
        self,
        vocab,
        num_classes,
        dim_token_emb,
        pos_emb_capacity,
        dim_pos,
        num_filters,
        kernel_size,
        dropout: Optional[float] = 0.5,
    ):
        super().__init__()

        self.token_embedding = StaticEmbedding(
            vocab.size, dim_token_emb, dropout=dropout
        )
        self.pos1_embedding = nn.Embedding(
            num_embeddings=pos_emb_capacity, embedding_dim=dim_pos
        )
        self.pos2_embedding = nn.Embedding(
            num_embeddings=pos_emb_capacity, embedding_dim=dim_pos
        )
        self.pcnn = PiecewiseCNN(
            dim_token_emb + 2 * dim_pos, num_filters, kernel_size, dropout=dropout
        )
        self.dense = nn.Linear(
            in_features=num_filters * 3, out_features=num_classes, bias=True
        )
        self.dropout = nn.Dropout(dropout)

    def forward(self, token_ids, head_pos, tail_pos, mask, scopes, labels=None):
        token_embedding = self.token_embedding(token_ids)
        pos1_embedding = self.pos1_embedding(head_pos)
        pos2_embedding = self.pos2_embedding(tail_pos)
        x = torch.cat([token_embedding, pos1_embedding, pos2_embedding], dim=-1)
        x = self.pcnn(x, mask)

        if labels is not None:
            bag_rep_for_loss = []
        bag_rep_for_pred = []

        # for training loss calculation
        if labels is not None:
            query = torch.zeros((x.size(0)), device=x.device, dtype=torch.long)
            for i in range(len(scopes)):
                query[scopes[i][0] : scopes[i][1]] = labels[i]
            att_mat = self.dense.weight[query]
            att_score = (x * att_mat).sum(-1)
            bag_rep_for_loss = []
            for i in range(len(scopes)):
                bag_mat = x[scopes[i][0] : scopes[i][1]]
                softmax_att_score = F.softmax(
                    att_score[scopes[i][0] : scopes[i][1]], dim=-1
                )
                bag_rep_for_loss.append(
                    (softmax_att_score.unsqueeze(-1) * bag_mat).sum(0)
                )
            bag_rep_for_loss = torch.stack(bag_rep_for_loss, 0)
            bag_rep_for_loss = self.dropout(bag_rep_for_loss)
            bag_rep_for_loss = self.dense(bag_rep_for_loss)

        # for prediction
        att_score = torch.matmul(x, self.dense.weight.transpose(0, 1))
        for i in range(len(scopes)):
            bag_mat = x[scopes[i][0] : scopes[i][1]]
            softmax_att_score = F.softmax(
                att_score[scopes[i][0] : scopes[i][1]].transpose(0, 1), dim=-1
            )
            rep_for_each_rel = torch.matmul(softmax_att_score, bag_mat)
            logit_for_each_rel = F.softmax(self.dense(rep_for_each_rel), dim=-1)
            logit_for_each_rel = logit_for_each_rel.diag()
            bag_rep_for_pred.append(logit_for_each_rel)
        bag_rep_for_pred = torch.stack(bag_rep_for_pred, 0)

        result = {}
        if labels is not None:
            result.update({"loss": F.cross_entropy(bag_rep_for_loss, labels)})
        result.update({"pred": bag_rep_for_pred.argmax(dim=-1)})

        return result



class Instance(object):
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return "Instance(" + str(self.__dict__) + ")"



from pathlib import Path
from typing import List, Optional, Tuple

import torch
from omegaconf import OmegaConf

from rex.data.data_manager import DataManager
from rex.data.transforms.base import TransformBase
from rex.tasks.base_task import TaskBase
from rex.utils.dict import get_dict_content
from rex.utils.io import dump_json
from rex.utils.progress_bar import pbar, rbar
from rex.utils.registry import register
from rex.utils.wrapper import safe_try


@register("task")
class SimpleTask(TaskBase):
    """Simple task with one model, one optimizer and (maybe) one lr scheduler"""

    def __init__(
        self,
        config: OmegaConf,
        initialize: Optional[bool] = True,
        makedirs: Optional[bool] = True,
        dump_configfile: Optional[bool] = True,
    ) -> None:
        super().__init__(config, initialize, makedirs, dump_configfile)

        self.middle_path = Path(config.task_dir).joinpath("middle")
        self.middle_path.mkdir(parents=True, exist_ok=True)
        self.measures_path = Path(config.task_dir).joinpath("measures")
        self.measures_path.mkdir(parents=True, exist_ok=True)

        self.logging("Init transform", level="DEBUG")
        self.transform: TransformBase = self.init_transform()
        self.logging("Init data_manager", level="DEBUG")
        self.data_manager: DataManager = self.init_data_manager()

        self.logging("Init model", level="DEBUG")
        self.model = self.init_model()

        if self.config.skip_train:
            self.optimizer = None
            self.lr_scheduler = None
        else:
            self.logging("Init optimizer", level="DEBUG")
            self.optimizer = self.init_optimizer()
            self.logging("Init lr_scheduler", level="DEBUG")
            self.lr_scheduler = self.init_lr_scheduler()

        self.after_initialized()
        self.logging("SimpleTask initialised", level="DEBUG")

    def after_initialized(self):
        self.logging("Prepare model", level="DEBUG")
        self.model = self.accelerator.prepare(self.model)
        self.logging("Prepare optimizer", level="DEBUG")
        self.optimizer = self.accelerator.prepare(self.optimizer)
        if self.lr_scheduler is not None:
            self.logging("Prepare lr_scheduler", level="DEBUG")
            self.lr_scheduler = self.accelerator.prepare(self.lr_scheduler)

    def init_transform(self):
        raise NotImplementedError

    def init_data_manager(self):
        raise NotImplementedError

    def init_model(self):
        raise NotImplementedError

    def init_optimizer(self):
        raise NotImplementedError

    def init_lr_scheduler(self):
        return None

    @torch.no_grad()
    def _get_eval_results_impl(
        self, input_batches: List, output_batches: List, *args, **kwargs
    ) -> dict:
        return self.get_eval_results(input_batches, output_batches, *args, **kwargs)

    def get_eval_results(
        self, input_batches: List, output_batches: List, *args, **kwargs
    ) -> dict:
        """Get evaluation measurements

        Args:
            input_batches: list of model input. Raw batch input.
            output_batches: list of model results (`preds`)

        Returns:
            ``rex.utils.dict.PrettyPrintDefaultDict``
                and ``rex.utils.dict.PrettyPrintDict`` is highly recommended
                to replace vanilla ``defaultdict`` and ``dict`` here.
        """
        raise NotImplementedError

    @torch.no_grad()
    def predict(self, *args, **kwargs):
        self.model.eval()
        return self.predict_api(*args, **kwargs)

    def predict_api(self, *args, **kwargs):
        raise NotImplementedError

    def get_data_loader(self, dataset_name, is_eval="guessing", epoch=0):
        loader = self.data_manager.load_loader(
            dataset_name, is_eval=is_eval, epoch=epoch
        )
        loader = self.accelerator.prepare(loader)
        return loader

    @safe_try
    def train(self):
        if self.config.skip_train:
            raise RuntimeError(
                "Training procedure started while config.skip_train is True!"
            )

        continue_training = True
        resumed_training = self.config.resumed_training
        train_loader = self.get_data_loader("train", False, 0)
        total_steps = self.history["curr_epoch"] * len(train_loader)
        for epoch_idx in rbar(
            self.history["curr_epoch"], self.config.num_epochs, desc="Epoch"
        ):
            if not resumed_training:
                self.history["curr_epoch"] = epoch_idx
            self.logging(f"Epoch: {epoch_idx}/{self.config.num_epochs}")

            self.model.train()
            self.optimizer.zero_grad()
            epoch_train_loader = self.get_data_loader(
                "train", is_eval=False, epoch=epoch_idx
            )
            loader = pbar(epoch_train_loader, desc=f"Train(e{epoch_idx})")
            for batch_idx, batch in enumerate(loader):
                if not resumed_training:
                    self.history["curr_batch"] = batch_idx
                    self.history["total_steps"] = total_steps
                if resumed_training and total_steps < self.history["total_steps"]:
                    continue
                elif resumed_training and total_steps == self.history["total_steps"]:
                    resumed_training = False

                result = self.model(**batch)

                loss = result["loss"] / self.config.grad_accum_steps
                loader.set_postfix({"loss": loss.item()})
                self.history["current_train_loss"] += loss.item()
                self.accelerator.backward(loss)

                if self.config.max_grad_norm > 0:
                    self.accelerator.clip_grad_norm_(
                        self.model.parameters(), max_norm=self.config.max_grad_norm
                    )
                if (
                    batch_idx + 1 % self.config.grad_accum_steps == 0
                    or batch_idx + 1 == len(loader)
                ):
                    self.optimizer.step()
                    if self.lr_scheduler is not None:
                        self.lr_scheduler.step()
                    self.optimizer.zero_grad()

                if (
                    self.config.step_eval_interval > 0
                    and total_steps + 1 % self.config.step_eval_interval
                ):
                    self._eval_during_train("step")
                    if not self._check_patience():
                        break
                total_steps += 1

            self.logging(loader)
            if not continue_training:
                # if break from inner step loop, early break
                break
            if (self.config.epoch_eval_interval > 0) and (
                ((epoch_idx + 1) % self.config.epoch_eval_interval) == 0
            ):
                self._eval_during_train("epoch")
                if not self._check_patience():
                    break

        self.logging("Trial finished.")
        if self.config.select_best_by_key == "metric":
            tmp_string = f"{self.history['best_metric']:.5f}"
        else:
            tmp_string = f"{self.history['best_loss']}"
        self.logging(
            f"Best epoch: {self.history['best_epoch']}, step: {self.history['best_step']}"
        )
        self.logging(
            f"Best {self.config.select_best_on_data}.{self.config.select_best_by_key}.{self.config.best_metric_field} : {tmp_string}"
        )

        if self.config.final_eval_on_test:
            self.logging("Loading best ckpt")
            self.load_best_ckpt()
            test_loss, test_measures = self.eval(
                "test", verbose=True, dump=True, postfix="final"
            )

    def _eval_during_train(self, eval_on: Optional[str] = "epoch"):
        """Evaluation during training to record eval info, and control training process

        Args:
            eval_on: epoch or step
        """
        # validate
        assert eval_on in [
            "epoch",
            "step",
        ], f"eval_on: {eval_on} must eq `epoch` or `step`"
        assert self.config.select_best_by_key in [
            "metric",
            "loss",
        ], f"select_best_by_key is {self.config.select_best_by_key}, while candidates are: `metric` and `loss`"
        assert (self.config.select_best_by_key == "train") or (
            self.config.select_best_by_key != "train"
            and self.config.select_best_on_data in self.config.eval_on_data
        ), f"{self.config.select_best_on_data} is not included in eval_on_data: {self.config.eval_on_data}"

        if len(self.config.eval_on_data) < 1:
            self.logging(
                "Does not provide any data to evaluate during training, continue training",
                level="WARNING",
            )
            return True

        # init
        curr_epoch_idx = self.history["curr_epoch"]
        curr_total_steps = self.history["total_steps"]
        history_index = curr_epoch_idx if eval_on == "epoch" else curr_total_steps
        history_index_identifier = f"{eval_on}.{history_index}"
        this_eval_result = {}  # to dump results

        # eval to get measurements and loss
        eval_on_datasets = set()
        for dataset_name in self.config.eval_on_data:
            dataset_name = DataManager._get_normalized_dataset_name(dataset_name)
            eval_on_datasets.add(dataset_name)
        for dataset_name in eval_on_datasets:
            eval_loss, eval_measures = self.eval(
                dataset_name, verbose=False, postfix=history_index_identifier
            )
            self.history[eval_on][dataset_name]["metrics"][
                history_index
            ] = eval_measures
            self.history[eval_on][dataset_name]["loss"][history_index] = eval_loss
            this_eval_result[f"{eval_on}.{dataset_name}.metrics"] = eval_measures
            this_eval_result[f"{eval_on}.{dataset_name}.loss"] = eval_loss

        self.history[eval_on][dataset_name]["loss"][history_index] = self.history[
            "current_train_loss"
        ]
        this_eval_result["train_loss"] = self.history["current_train_loss"]

        # update the best
        select_best_on_data = DataManager._get_normalized_dataset_name(
            self.config.select_best_on_data
        )
        metric = self.history[eval_on][select_best_on_data]["metrics"][history_index]
        metric = get_dict_content(metric, self.config.best_metric_field)
        is_best_metric = False
        if metric > self.history["best_metric"]:
            is_best_metric = True
            self.history["best_metric"] = metric
        this_eval_result["is_best_metric"] = is_best_metric

        loss = self.history[eval_on][select_best_on_data]["loss"][history_index]
        is_best_loss = False
        if loss < self.history["best_loss"]:
            is_best_loss = True
            self.history["best_loss"] = loss
        this_eval_result["is_best_loss"] = is_best_loss

        # count no climbing
        is_best = False
        if (self.config.select_best_by_key == "metric" and is_best_metric) or (
            self.config.select_best_by_key == "loss" and is_best_loss
        ):
            is_best = True
            self.history["best_epoch"] = curr_epoch_idx
            self.history["best_step"] = curr_total_steps
            if eval_on == "epoch":
                self.history["no_climbing_epoch_cnt"] = 0
            else:
                self.history["no_climbing_step_cnt"] = 0
        else:
            if eval_on == "epoch":
                self.history["no_climbing_epoch_cnt"] += self.config.epoch_eval_interval
            else:
                self.history["no_climbing_step_cnt"] += self.config.step_eval_interval
        this_eval_result["is_best"] = is_best
        this_eval_result["no_climbing_epoch_cnt"] = self.history[
            "no_climbing_epoch_cnt"
        ]
        this_eval_result["no_climbing_step_cnt"] = self.history["no_climbing_step_cnt"]

        # print results
        self.logging(
            f"Eval on {eval_on}, Idx: {history_index_identifier}, is_best: {is_best}"
        )
        self.logging(f"Train loss: {self.history['current_train_loss']:.5f}")
        for dataset_name in eval_on_datasets:
            eval_measures = self.history[eval_on][dataset_name]["metrics"][
                history_index
            ]
            eval_loss = self.history[eval_on][dataset_name]["loss"][history_index]
            self.logging(
                (
                    f"{dataset_name} - "
                    f"{self.config.best_metric_field}: "
                    f"{get_dict_content(eval_measures, self.config.best_metric_field):.5f}, "
                    f"eval loss: {eval_loss:.5f}"
                )
            )
        if self.config.select_best_by_key == "metric":
            tmp_string = f"{self.history['best_metric']:.5f}"
        else:
            tmp_string = f"{self.history['best_loss']:.5f}"
        self.logging(f"Best {self.config.select_best_by_key}: {tmp_string}")
        self.logging(f"Best {eval_on}: {self.history[f'best_{eval_on}']}")
        self.logging(
            f"No Climbing Count of {eval_on}: {self.history[f'no_climbing_{eval_on}_cnt']}"
        )
        dump_json(
            this_eval_result,
            self.measures_path.joinpath(f"{history_index_identifier}.json"),
            indent=2,
        )

        # reset current training loss
        self.history["current_train_loss"] = 0.0

        # save checkpoints
        if self.config.save_every_ckpt:
            self.save_ckpt(f"{history_index_identifier}")
        if is_best and self.config.save_best_ckpt:
            self.save_ckpt("best")

    def _check_patience(self):
        """Check patience, returns False if training process should be stopped, else returns True"""
        if (
            self.config.epoch_patience > 0
            and self.history["no_climbing_epoch_cnt"] >= self.config.epoch_patience
        ) or (
            self.config.step_patience > 0
            and self.history["no_climbing_step_cnt"] >= self.config.step_patience
        ):
            self.logging(
                (
                    "Early Stopped: No climbing count: "
                    f"Epoch: {self.history['no_climbing_epoch_cnt']} / {self.config.epoch_patience} "
                    f"Step: {self.history['no_climbing_step_cnt']} / {self.config.step_patience} "
                )
            )
            return False
        if (
            self.config.num_steps > 0
            and self.history["total_steps"] >= self.config.num_steps
        ):
            self.logging(
                f"Reached the max num of steps: {self.history['total_steps']} / {self.config.num_steps}"
            )
            return False
        return True

    @torch.no_grad()
    def eval(
        self, dataset_name, verbose=False, dump=False, postfix=""
    ) -> Tuple[float, dict]:
        """Eval on specific dataset and return loss and measurements

        Args:
            dataset_name: which dataset to evaluate
            verbose: whether to log evaluation results
            dump: if True, dump result to this filepath
            postfix: filepath postfix for dumping

        Returns:
            eval_loss: float
            metrics: dict
        """
        self.model.eval()
        eval_loader = self.get_data_loader(
            dataset_name, is_eval=True, epoch=self.history["curr_epoch"]
        )
        loader = pbar(
            eval_loader, desc=f"{dataset_name} - {postfix} Eval", ncols=80, ascii=True
        )

        eval_loss = 0.0
        metrics = {}
        origin = []
        output = []
        # raw_batch: dict
        for batch in loader:
            out = self.model(**batch)
            eval_loss += out["loss"].item()
            all_batch = self.accelerator.gather(batch)
            all_out = self.accelerator.gather(out)
            origin.append(all_batch)
            output.append(all_out)

        metrics = self._get_eval_results_impl(origin, output)

        if verbose:
            self.logging(f"Eval dataset: {dataset_name}")
            self.logging(f"Eval loss: {eval_loss}")
            self.logging(
                f"Eval metrics: {get_dict_content(metrics, self.config.best_metric_field)}"
            )
        if dump:
            dump_obj = {
                "dataset_name": dataset_name,
                "eval_loss": eval_loss,
                "metrics": metrics,
            }
            dump_json(
                dump_obj, self.measures_path.joinpath(f"{dataset_name}.{postfix}.json")
            )

        return eval_loss, metrics



import math
from typing import Callable, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


class PiecewiseCNN(nn.Module):
    def __init__(
        self,
        in_channels,
        num_filters,
        kernel_size,
        dropout: Optional[float] = 0.5,
        activation_function: Optional[Callable] = torch.tanh,
    ):
        super().__init__()

        self.conv = nn.Conv1d(
            in_channels,
            num_filters,
            kernel_size,
            padding=math.floor((kernel_size - 1) / 2),
        )
        self.dropout = nn.Dropout(dropout)
        self.act = activation_function

        # mask operation for pcnn
        masks = torch.tensor(
            [[0, 0, 0], [100, 0, 0], [0, 100, 0], [0, 0, 100]], dtype=torch.float
        )
        self.mask_embedding = nn.Embedding.from_pretrained(masks, freeze=True)

    def forward(self, input_rep, mask):
        # mask: (batch_size, seq_len)
        # input_rep: (batch_size, seq_len, hidden_size) -> (batch_size, hidden_size, seq_len)
        x = input_rep.permute(0, 2, 1)
        # x: (batch_size, num_filters, seq_len)
        x = self.conv(x)

        # mask_embed: (batch_size, seq_len, 3)
        mask_embed = self.mask_embedding(mask)
        # mask_embed: (batch_size, 1, seq_len, 3)
        mask_embed = mask_embed.unsqueeze(dim=1)
        # x: (batch_size, num_filters, seq_len, 3)
        x = x.unsqueeze(-1) + mask_embed
        # x: (batch_size, num_filters, 3)
        x = torch.max(x, dim=2)[0] - 100
        # x: (batch_size, 3*num_filters)
        x = x.view(x.shape[0], -1)

        x = torch.tanh(x)
        x = self.dropout(x)
        return x



from typing import List, Tuple


def get_entities_from_tag_seq(chars: List[str], tags: List[str]) -> List[Tuple]:
    r"""
    get entities from a seqence of chars and tags, BIO and BMES schemas are both supported
    Args:
        chars: list of chars (single string)
        tags: list of tags (in string format), like `B`, `I-Subj`, `B_PER`.
            if there is no postfix type string, entities will not be categorized
    Returns:
        list of tuples:
            [(entity name, entity type, start position, end position + 1), ...]
    """
    if len(tags) > len(chars):
        tags = tags[: len(chars)]
    elif len(chars) > len(tags):
        chars = chars[: len(tags)]
    entity_list = []
    last_type = ""
    ent = ""
    ent_start = -1

    for idx, (char, tag) in enumerate(zip(chars, tags)):
        if len(tag) > 2:
            curr_type = tag[2:]
        else:
            curr_type = "default"

        if tag.startswith("B"):
            if len(ent) > 0:
                entity_list.append((ent, last_type, ent_start, idx))

            ent = char
            last_type = curr_type
            ent_start = idx

        elif tag.startswith("I") or tag.startswith("M"):
            if curr_type == last_type:
                ent += char
            else:
                # illegal case, early stop here
                if len(ent) > 0:
                    entity_list.append((ent, last_type, ent_start, idx))
                ent = ""
                last_type = ""
                ent_start = -1

        elif tag.startswith("E"):
            if curr_type == last_type:
                ent += char

            if len(ent) > 0:
                entity_list.append((ent, last_type, ent_start, idx + 1))

            ent = ""
            last_type = ""
            ent_start = -1

        elif tag.startswith("S"):
            if len(ent) > 0:
                # last entity is not stopped
                entity_list.append((ent, last_type, ent_start, idx))

            entity_list.append((char, curr_type, idx, idx + 1))

        else:
            # O
            if len(ent) > 0:
                entity_list.append((ent, last_type, ent_start, idx))

            ent = ""
            last_type = ""
            ent_start = -1

    if len(ent) > 0:
        entity_list.append((ent, last_type, ent_start, ent_start + len(ent)))

    return entity_list


def get_num_illegal_tags_from_tag_seq(tags: List[str]) -> int:
    r"""
    get number of illegal tags from a seqence of chars and tags, BIO and BMES schemas are both supported
    Args:
        tags: list of tags (in string format), like `B`, `I-Subj`, `B_PER`.
            if there is no postfix type string, entities will not be categorized
    Returns:
        number of illegal tags
    """
    num_illegal_tags = 0
    last_type = "default"
    last_tag = "O"

    for tag in tags:
        if len(tag) > 2:
            curr_type = tag[2:]
        else:
            curr_type = "default"

        if tag[0] in "IME":
            if last_tag not in "BIM" or curr_type != last_type:
                # illegal case
                num_illegal_tags += 1
        elif tag[0] == "S" and last_tag in "IM":
            num_illegal_tags += 1

        last_type = curr_type
        last_tag = tag[0]

    return num_illegal_tags



import os
import random
from typing import Optional

import numpy as np
import torch
from omegaconf.omegaconf import OmegaConf

from rex.tasks import CONFIG_PARAMS_FILENAME, LOG_FILENAME
from rex.utils.logging import logger
from rex.utils.progress_bar import pbar


def set_seed(seed: Optional[int] = 1227, set_cudnn: Optional[bool] = False):
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)

    if set_cudnn:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def set_seed_and_log_path(
    seed: Optional[int] = 1227,
    log_path: Optional[str] = None,
    set_cudnn: Optional[bool] = False,
):
    set_seed(seed, set_cudnn=set_cudnn)
    if log_path:
        logger.add(log_path, backtrace=True, diagnose=True)


def set_device(device):
    torch.cuda.set_device(device)


def init_all(
    task_dir: str,
    seed: Optional[int] = 1227,
    set_cudnn: Optional[bool] = False,
    config: Optional[OmegaConf] = None,
):
    set_seed(seed, set_cudnn=set_cudnn)
    logger_handler_id = logger.add(
        os.path.join(task_dir, LOG_FILENAME), backtrace=True, diagnose=True
    )
    if config is not None:
        OmegaConf.save(
            config, os.path.join(task_dir, CONFIG_PARAMS_FILENAME), resolve=True
        )
    return logger_handler_id



from typing import Optional, Union

import torch


def move_to_cuda_device(
    obj: Union[torch.Tensor, dict], device: Optional[torch.device] = None
):
    if isinstance(obj, torch.Tensor):
        obj = obj.cuda(device)
    elif isinstance(obj, dict):
        for key in obj:
            obj[key] = move_to_cuda_device(obj[key], device)
    return obj


def move_to_device(
    obj: Union[torch.Tensor, dict], device: Optional[torch.device] = None
):
    if isinstance(obj, torch.Tensor):
        obj = obj.to(device)
    elif isinstance(obj, dict):
        for key in obj:
            obj[key] = move_to_device(obj[key], device)
    return obj


def detach_cpu_list(tensor: torch.Tensor):
    return tensor.detach().cpu().tolist()



import json
from collections import defaultdict


def get_dict_content(dict_item: dict, key: str):
    val = dict_item
    for k in key.split("."):
        val = val[k]
    return val


def _pretty_format(
    obj: dict, decimal=3, return_percentage=False, add_percentage_symbol=True
):
    if not isinstance(obj, dict):
        return obj

    pretty = {}
    for key, val in obj.items():
        if isinstance(val, float):
            new_val = val
            if return_percentage:
                new_val = 100 * new_val
            if add_percentage_symbol:
                template = f"{{:.{decimal}f}} %"
            else:
                template = f"{{:.{decimal}f}}"
            new_val = template.format(new_val)
        elif isinstance(val, dict):
            new_val = _pretty_format(
                val,
                decimal=decimal,
                return_percentage=return_percentage,
                add_percentage_symbol=add_percentage_symbol,
            )
        else:
            new_val = val
        pretty[key] = new_val
    return pretty


class PrettyPrintDict(dict):
    def __str__(
        self, indent=2, decimal=3, return_percentage=True, add_percentage_symbol=True
    ):
        pretty_dict = _pretty_format(
            self,
            decimal=decimal,
            return_percentage=return_percentage,
            add_percentage_symbol=add_percentage_symbol,
        )
        return json.dumps(pretty_dict, ensure_ascii=False, indent=indent)

    def to_dict(self) -> dict:
        return dict(self)

    def jsonify(self, **kwargs) -> str:
        return json.dumps(dict(self), ensure_ascii=False, **kwargs)


class PrettyPrintDefaultDict(defaultdict):
    def __str__(
        self, indent=2, decimal=3, return_percentage=True, add_percentage_symbol=True
    ):
        pretty_dict = _pretty_format(
            self,
            decimal=decimal,
            return_percentage=return_percentage,
            add_percentage_symbol=add_percentage_symbol,
        )
        return json.dumps(pretty_dict, ensure_ascii=False, indent=indent)

    def to_dict(self) -> dict:
        return dict(self)

    def jsonify(self, **kwargs) -> str:
        return json.dumps(dict(self), ensure_ascii=False, **kwargs)



from rex.data.vocab import Vocab
from rex.utils.config import ConfigArgument, ConfigParser
from rex.utils.io import load_embedding_file, load_line_iterator
from rex.utils.logging import logger
from rex.utils.progress_bar import pbar


def build_emb(raw_emb_filepath, dump_emb_filepath, *files):
    """build vocab and filter out useless embeddings from ``raw_emb_filepath``"""
    logger.info(f"loading embedding from {raw_emb_filepath}")
    tokens, token2vec = load_embedding_file(raw_emb_filepath)

    logger.info("building vocab")
    raw_vocab = set()
    for filepath in pbar(files):
        for line in load_line_iterator(filepath):
            raw_vocab.update(line.strip())

    logger.info("updating vocab")
    final_tokens = raw_vocab & set(tokens)
    final_vocab = Vocab(init_pad_unk_emb=True)
    for token in pbar(final_tokens):
        final_vocab.add(token, token2vec[token])

    logger.info("dumping vocab")
    final_vocab.save_pretrained(dump_emb_filepath, dump_weights=True)


if __name__ == "__main__":
    config = ConfigParser.parse_cmd(
        ConfigArgument("-r", "--raw-emb-filepath", help="raw embedding filepath"),
        ConfigArgument("-o", "--dump-emb-filepath", help="dumped embedding filepath"),
        ConfigArgument("-f", "--filepaths", nargs="*", help="raw text filepaths"),
    )
    build_emb(config.raw_emb_filepath, config.dump_emb_filepath, *config.filepaths)



from typing import Iterable, Iterator, Optional

from torch.utils.data import Dataset, IterableDataset


class CachedDataset(Dataset):
    def __init__(self, data: Iterable) -> None:
        super().__init__()
        self.data = data

    def __getitem__(self, index: int):
        return self.data[index]

    def __len__(self) -> int:
        return len(self.data)


class CachedBagREDataset(Dataset):
    def __init__(self, data_with_scopes) -> None:
        super().__init__()
        data, scopes = data_with_scopes
        self.data = data
        self.scopes = scopes

    def __getitem__(self, index: int):
        results = []
        for idx in self.scopes[index]:
            results.append(self.data[idx])
        return results

    def __len__(self) -> int:
        return len(self.scopes)


class StreamTransformDataset(Dataset):
    def __init__(
        self, data: Iterable, transform, debug: Optional[bool] = False
    ) -> None:
        super().__init__()
        if debug:
            data = data[:128]
        self.data = data
        self.transform = transform

    def __getitem__(self, index: int):
        return self.transform(self.data[index])

    def __len__(self) -> int:
        return len(self.data)


class StreamReadDataset(IterableDataset):
    def __init__(
        self, data_iterator: Iterator, transform, debug: Optional[bool] = False
    ) -> None:
        super().__init__()

        self.transform = transform
        self.data_iterator = data_iterator
        self.debug = debug
        self.cnt = 0

    def __iter__(self):
        for item in self.data_iterator:
            if self.debug and self.cnt >= 500:
                raise StopIteration
            yield self.transform(item)
            self.cnt += 1



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from transformers.models.bert import BertModel

from rex.modules.embeddings.static_embedding import StaticEmbedding
from rex.modules.span import SubjObjSpan


class LSTMCasRel(nn.Module):
    def __init__(
        self,
        vocab,
        num_lstm_layers,
        dim_token_emb,
        num_classes,
        pred_threshold: Optional[float] = 0.5,
        dropout: Optional[float] = 0.5,
    ):
        super().__init__()
        if dim_token_emb % 2 != 0:
            raise ValueError("Dimension of ``dim_token_emb`` must be odd")

        self.token_embedding = StaticEmbedding(
            vocab.size, dim_token_emb, dropout=dropout
        )
        self.encoder = nn.LSTM(
            dim_token_emb,
            dim_token_emb // 2,
            num_lstm_layers,
            bias=True,
            batch_first=True,
            dropout=dropout,
            bidirectional=True,
        )
        self.span_nn = SubjObjSpan(dim_token_emb, num_classes, pred_threshold)

    def encode(self, hidden, mask):
        batch_size, seq_len, hidden_size = hidden.shape
        total_length = mask.shape[1]
        lens = mask.sum(dim=1)

        x = pack_padded_sequence(
            hidden, lens.cpu(), batch_first=True, enforce_sorted=False
        )
        output, (_, _) = self.encoder(x)  # h_n: #(direction, #sent, hidden_size)
        output, _ = pad_packed_sequence(
            output, batch_first=True, padding_value=0.0, total_length=total_length
        )
        return output

    def forward(
        self,
        token_ids,
        mask,
        subj_heads=None,
        subj_tails=None,
        subj_head=None,
        subj_tail=None,
        obj_head=None,
        obj_tail=None,
    ):
        result = {}
        emb = self.token_embedding(token_ids)
        hidden = self.encode(emb, mask)
        # hidden = emb

        if self.training:
            (
                subj_head_logits,
                subj_tail_logits,
                obj_head_logits,
                obj_tail_logits,
            ) = self.span_nn(hidden, subj_head, subj_tail)

            subj_head_loss = F.binary_cross_entropy_with_logits(
                subj_head_logits, subj_heads, reduction="none"
            )
            subj_head_loss = (subj_head_loss * mask).sum() / mask.sum()
            subj_tail_loss = F.binary_cross_entropy_with_logits(
                subj_tail_logits, subj_tails, reduction="none"
            )
            subj_tail_loss = (subj_tail_loss * mask).sum() / mask.sum()
            subj_loss = subj_head_loss + subj_tail_loss

            result["subj_loss"] = subj_loss
            result["subj_head_loss"] = subj_head_loss
            result["subj_tail_loss"] = subj_tail_loss

            obj_head_loss = F.binary_cross_entropy_with_logits(
                obj_head_logits, obj_head, reduction="none"
            )
            obj_head_loss = (obj_head_loss * mask.unsqueeze(-1)).sum() / mask.sum()
            obj_tail_loss = F.binary_cross_entropy_with_logits(
                obj_tail_logits, obj_tail, reduction="none"
            )
            obj_tail_loss = (obj_tail_loss * mask.unsqueeze(-1)).sum() / mask.sum()
            obj_loss = obj_head_loss + obj_tail_loss

            result["obj_loss"] = obj_loss
            result["obj_head_loss"] = obj_head_loss
            result["obj_tail_loss"] = obj_tail_loss

            result["loss"] = subj_loss + obj_loss
        else:
            triples = self.span_nn.predict(hidden)
            result["pred"] = triples

        return result


class CasRel(nn.Module):
    def __init__(
        self,
        bert_model_dir,
        num_classes,
        pred_threshold: Optional[float] = 0.5,
    ):
        super().__init__()
        self.pred_threshold = pred_threshold

        self.bert_encoder = BertModel.from_pretrained(bert_model_dir)
        self.hidden_size = self.bert_encoder.config.hidden_size
        self.span_nn = SubjObjSpan(self.hidden_size, num_classes, pred_threshold)

    def encode(self, input_ids, attention_mask):
        outputs = self.bert_encoder(input_ids=input_ids, attention_mask=attention_mask)
        return outputs.last_hidden_state

    def forward(
        self,
        token_ids,
        mask,
        subj_heads=None,
        subj_tails=None,
        subj_head=None,
        subj_tail=None,
        obj_head=None,
        obj_tail=None,
    ):
        result = {}
        hidden = self.encode(token_ids, mask)

        if self.training:
            (
                subj_head_logits,
                subj_tail_logits,
                obj_head_logits,
                obj_tail_logits,
            ) = self.span_nn(hidden, subj_head, subj_tail)

            subj_head_loss = F.binary_cross_entropy_with_logits(
                subj_head_logits, subj_heads, reduction="none"
            )
            subj_head_loss = (subj_head_loss * mask).sum() / mask.sum()
            subj_tail_loss = F.binary_cross_entropy_with_logits(
                subj_tail_logits, subj_tails, reduction="none"
            )
            subj_tail_loss = (subj_tail_loss * mask).sum() / mask.sum()
            subj_loss = subj_head_loss + subj_tail_loss

            result["subj_loss"] = subj_loss
            result["subj_head_loss"] = subj_head_loss
            result["subj_tail_loss"] = subj_tail_loss

            obj_head_loss = F.binary_cross_entropy_with_logits(
                obj_head_logits, obj_head, reduction="none"
            )
            obj_head_loss = (obj_head_loss * mask.unsqueeze(-1)).sum() / mask.sum()
            obj_tail_loss = F.binary_cross_entropy_with_logits(
                obj_tail_logits, obj_tail, reduction="none"
            )
            obj_tail_loss = (obj_tail_loss * mask.unsqueeze(-1)).sum() / mask.sum()
            obj_loss = obj_head_loss + obj_tail_loss

            result["obj_loss"] = obj_loss
            result["obj_head_loss"] = obj_head_loss
            result["obj_tail_loss"] = obj_tail_loss

            result["loss"] = subj_loss + obj_loss
        else:
            triples = self.span_nn.predict(hidden)
            result["pred"] = triples

        return result



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from rex.modules.ffn import MLP
from rex.utils.span import find_closest_span_pairs, find_closest_span_pairs_with_index


class SubjObjSpan(nn.Module):
    """
    Inputs:
        hidden: (batch_size, seq_len, hidden_size)
        one_subj_head: object golden head with one subject (batch_size, hidden_size)
        one_subj_tail: object golden tail with one subject (batch_size, hidden_size)
    """

    def __init__(self, hidden_size, num_classes, threshold: Optional[float] = 0.5):
        super().__init__()
        self.threshold = threshold
        self.subj_head_ffnn = nn.Linear(hidden_size, 1)
        self.subj_tail_ffnn = nn.Linear(hidden_size, 1)
        self.obj_head_ffnn = nn.Linear(hidden_size, num_classes)
        self.obj_tail_ffnn = nn.Linear(hidden_size, num_classes)

    def get_objs_for_specific_subj(self, subj_head_mapping, subj_tail_mapping, hidden):
        # (batch_size, 1, hidden_size)
        subj_head = torch.matmul(subj_head_mapping, hidden)
        # (batch_size, 1, hidden_size)
        subj_tail = torch.matmul(subj_tail_mapping, hidden)
        # (batch_size, 1, hidden_size)
        sub = (subj_head + subj_tail) / 2
        # (batch_size, seq_len, hidden_size)
        encoded_text = hidden + sub
        # (batch_size, seq_len, num_classes)
        pred_obj_heads = self.obj_head_ffnn(encoded_text)
        # (batch_size, seq_len, num_classes)
        pred_obj_tails = self.obj_tail_ffnn(encoded_text)
        return pred_obj_heads, pred_obj_tails

    def build_mapping(self, subj_heads, subj_tails):
        """
        Build head & tail mapping for predicted subjects,
        for each instance in a batch, for a subject in all
        the predicted subjects, return a single subject
        and its corresponding mappings.
        """
        for subj_head, subj_tail in zip(subj_heads, subj_tails):
            subjs = find_closest_span_pairs(subj_head, subj_tail)
            seq_len = subj_head.shape[0]
            for subj in subjs:
                subj_head_mapping = torch.zeros(seq_len, device=subj_head.device)
                subj_tail_mapping = torch.zeros(seq_len, device=subj_tail.device)
                subj_head_mapping[subj[0]] = 1.0
                subj_tail_mapping[subj[1]] = 1.0
                yield subj, subj_head_mapping, subj_tail_mapping

    def build_batch_mapping(self, subj_head, subj_tail):
        """
        Build head & tail mapping for predicted subjects,
        for each instance in a batch, return all the predicted
        subjects and mappings.
        """
        subjs = find_closest_span_pairs(subj_head, subj_tail)
        seq_len = subj_head.shape[0]
        if len(subjs) > 0:
            subjs_head_mapping = torch.zeros(
                len(subjs), seq_len, device=subj_head.device
            )
            subjs_tail_mapping = torch.zeros(
                len(subjs), seq_len, device=subj_tail.device
            )

            for subj_idx, subj in enumerate(subjs):
                subjs_head_mapping[subj_idx, subj[0]] = 1.0
                subjs_tail_mapping[subj_idx, subj[1]] = 1.0
            return subjs, subjs_head_mapping, subjs_tail_mapping
        else:
            return None, None, None

    def forward(self, hidden, subj_head, subj_tail):
        # subj_head_out, subj_tail_out: (batch_size, seq_len, 1)
        subj_head_out = self.subj_head_ffnn(hidden)
        subj_tail_out = self.subj_tail_ffnn(hidden)
        # subj_head, subj_tail: (batch_size, seq_len)
        # obj_head_out, obj_tail_out: (batch_size, seq_len, num_classes)
        obj_head_out, obj_tail_out = self.get_objs_for_specific_subj(
            subj_head.unsqueeze(1), subj_tail.unsqueeze(1), hidden
        )

        return (
            subj_head_out.squeeze(-1),
            subj_tail_out.squeeze(-1),
            obj_head_out,
            obj_tail_out,
        )

    def predict(self, hidden):
        # hidden: 1 x hidden_size
        if hidden.shape[0] != 1:
            raise RuntimeError(
                (
                    "eval batch size must be 1 x hidden_size, "
                    f"while hidden is {hidden.shape}"
                )
            )
        # 1, hidden_size, 1
        subj_head_out = self.subj_head_ffnn(hidden)
        # 1, hidden_size, 1
        subj_tail_out = self.subj_tail_ffnn(hidden)
        subj_head_out = torch.sigmoid(subj_head_out)
        subj_tail_out = torch.sigmoid(subj_tail_out)
        pred_subj_head = subj_head_out.ge(self.threshold).long()
        pred_subj_tail = subj_tail_out.ge(self.threshold).long()

        triples = []
        # subjs: [(1, 3), (3, 3), (17, 17)]
        # subj_head_mappings: (len(subjs), seq_len)
        subjs, subj_head_mappings, subj_tail_mappings = self.build_batch_mapping(
            pred_subj_head.squeeze(0).squeeze(-1), pred_subj_tail.squeeze(0).squeeze(-1)
        )

        if subjs:
            # obj_head_out: (len(subjs), seq_len, num_classes)
            obj_head_out, obj_tail_out = self.get_objs_for_specific_subj(
                subj_head_mappings.unsqueeze(1), subj_tail_mappings.unsqueeze(1), hidden
            )
            obj_head_out = torch.sigmoid(obj_head_out)
            obj_tail_out = torch.sigmoid(obj_tail_out)
            obj_head_out = obj_head_out.ge(self.threshold).long()
            obj_tail_out = obj_tail_out.ge(self.threshold).long()
            for subj_idx, subj in enumerate(subjs):
                objs = find_closest_span_pairs_with_index(
                    obj_head_out[subj_idx].permute(1, 0),
                    obj_tail_out[subj_idx].permute(1, 0),
                )
                for relation_idx, obj_pair_start, obj_pair_end in objs:
                    triples.append(
                        (
                            (subj[0], subj[1] + 1),
                            relation_idx,
                            (obj_pair_start, obj_pair_end + 1),
                        )
                    )
        return [triples]



import argparse
import importlib
import pathlib
from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import Iterable, List, Optional

from omegaconf import OmegaConf


class ConfigArgument(object):
    """Instance for storing config argument"""

    def __init__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs


class ConfigParser(argparse.ArgumentParser):
    """Configuration Parser

    Args:
        *args (~Iterable[ConfigArgument]): arguments to be added in ```add_argument``` function
        **kwargs: arguments of ```argparse.ArgumentParser```

    Examples:
        >>> config = ConfigParser.parse_cmd()
        >>> config = ConfigParser(
        ...     ConfigArgument('-d', '--task-dir', type=str, help='path to task directory'),
        ...     ConfigArgument('-n', '--task-name', type=str, help='name of task')
        ... ).parse_cmd()
    """

    def __init__(
        self,
        *args: Iterable[ConfigArgument],
        init_priority_args: Optional[bool] = True,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)

        for arg in args:
            assert isinstance(arg, ConfigArgument)
            self.add_argument(*arg.args, **arg.kwargs)

        self._inited = False
        if init_priority_args:
            self._inited = True
            self.init_priority_args()

    def init_priority_args(self):
        self.add_argument(
            "-d",
            "--use-default-base-config",
            default=False,
            action="store_true",
            help="whether to use internal base config",
        )
        self.add_argument(
            "-b",
            "--base-config-filepath",
            type=pathlib.Path,
            help="base configuration filepath, could be a template, lowest priority",
        )
        self.add_argument(
            "-c",
            "--custom-config-filepath",
            type=pathlib.Path,
            help="configuration filepath, customised configs, middle priority",
        )
        self.add_argument(
            "-a",
            "--additional-args",
            type=str,
            nargs="*",
            help="additional args in dot-list format, highest priority",
        )

    @classmethod
    def parse_cmd_args(cls, *args, cmd_args: Optional[List[str]] = None, **kwargs):
        parser = cls(*args, **kwargs)
        args = parser.parse_args(args=cmd_args)
        return parser, args

    @classmethod
    def parse_args_config(cls, parser, args):
        if parser._inited:
            for path in ["base_config_filepath", "custom_config_filepath"]:
                # arg: pathlib.Path
                arg = getattr(args, path)
                if arg is not None:
                    assert arg.exists() is True and arg.is_file() is True
                    setattr(args, path, str(arg.absolute()))

        config = parser.convert_args_to_config(args)

        return config

    @classmethod
    def parse_cmd(
        cls, *args, cmd_args: Optional[List[str]] = None, **kwargs
    ) -> OmegaConf:
        """
        Get command arguments from `base_config_filepath`, `custom_config_filepath` and `additional_args`.

        To make configurations flexible, configs are split into three levels:
            - `base_config_filepath`: path to basic configuration file in yaml format, lowest priority
            - `custom_config_filepath`: path to customised config file in yaml format, middle priority.
                If this file is loaded, it will override configs in the `base_config_filepath`.
            - `additional_args`: additional args in dot-list format, highest priority.
                If these args are set, it will override configs in the previous files.

        Args:
            *args: for initialise the config parser object
            cmd_args: if you don't want to use the command line, provide `cmd_args` like `sys.argv[1:]`
            **kwargs: for initialise the config parser object

        Examples:
            $ python run.py -d
            $ python run.py -b conf/config.yaml
            $ python run.py -c conf/re/sent_ipre.yaml
            $ python run.py -a lr.encoder=1e-4 dropout=0.6
            $ python run.py -b conf/config.yaml -c conf/re/sent_ipre.yaml
            $ python run.py -b conf/config.yaml -a lr.encoder=1e-4 dropout=0.6
            $ python run.py -c conf/re/sent_ipre.yaml -a lr.encoder=1e-4 dropout=0.6
            $ python run.py -b conf/config.yaml -c conf/re/sent_ipre.yaml -a lr.encoder=1e-4 dropout=0.6
        """
        parser, args = cls.parse_cmd_args(*args, cmd_args=cmd_args, **kwargs)
        config = cls.parse_args_config(parser, args)

        return config

    def convert_args_to_config(self, args: argparse.Namespace) -> OmegaConf:
        """Convert args into config"""
        config = OmegaConf.create(
            {"_config_info": {"create_time": datetime.now().strftime("%F %X")}}
        )
        OmegaConf.set_readonly(config, False)

        if "use_default_base_config" in args:
            config._config_info.use_default_base_config = args.use_default_base_config
            if args.use_default_base_config:
                default_config = DefaultBaseConfig()
                config.merge_with(asdict(default_config))

        if self._inited:
            base_config_filepath = args.base_config_filepath
            if base_config_filepath is not None:
                base_config = OmegaConf.load(base_config_filepath)
                config.merge_with(base_config)
                config._config_info.base_filepath = base_config_filepath

            custom_config_filepath = args.custom_config_filepath
            if custom_config_filepath is not None:
                custom_config = OmegaConf.load(custom_config_filepath)
                config.merge_with(custom_config)
                config._config_info.custom_config_filepath = custom_config_filepath

            additional_args = args.additional_args
            if additional_args is not None:
                add_args = OmegaConf.from_dotlist(additional_args)
                config.merge_with(add_args)
                config._config_info.additional_args = additional_args

        kwargs = dict(args._get_kwargs())
        for k in ["base_config_filepath", "custom_config_filepath", "additional_args"]:
            if k in kwargs:
                kwargs.pop(k)
        config.merge_with(kwargs)

        return config


@dataclass
class DefaultBaseConfig:
    # task
    task_name: str = field(
        default="temp_task",
        metadata={"help": "name of task, used for creating task directory"},
    )
    # filepaths
    output_dir: str = field(
        default="outputs", metadata={"help": "base output directory"}
    )
    task_dir: str = field(
        default="outputs/temp_task",
        metadata={"help": "task directory"},
    )
    data_dir: str = field(default="data", metadata={"help": "data directory"})
    train_filepath: str = field(
        default="train.jsonl", metadata={"help": "filepath to train set"}
    )
    dev_filepath: str = field(
        default="dev.jsonl", metadata={"help": "filepath to dev set"}
    )
    test_filepath: str = field(
        default="test.jsonl", metadata={"help": "filepath to test set"}
    )
    # training control
    device: str = field(default="cpu", metadata={"help": "device string"})
    random_seed: int = field(default=1227, metadata={"help": "random seed"})
    num_epochs: int = field(
        default=50, metadata={"help": "max number of training epochs"}
    )
    num_steps: int = field(
        default=-1,
        metadata={
            "help": "max number of training steps. `-1` if not training in steps"
        },
    )
    epoch_patience: int = field(
        default=5,
        metadata={"help": "stop training if the metric does not grow in [x] epochs"},
    )
    step_patience: int = field(
        default=5000,
        metadata={"help": "stop training if the metric does not grow in [x] steps"},
    )
    batch_size: int = field(
        default=64,
        metadata={
            "help": "training batch size for each process (if ddp, the real batch size is num_process * batch_size)"
        },
    )
    learning_rate: float = field(
        default=1e-3, metadata={"help": "training learning rate"}
    )
    max_grad_norm: int = field(
        default=-1, metadata={"help": "max gradient norm. `-1`: no clipping"}
    )
    skip_train: bool = field(
        default=False, metadata={"help": "whether to skip training process"}
    )
    debug_mode: bool = field(
        default=False,
        metadata={
            "help": "whether to enter debug mode (use less data for debug and test)"
        },
    )
    grad_accum_steps: int = field(
        default=1, metadata={"help": "gradient accumulation steps"}
    )
    resumed_training: bool = field(
        default=False, metadata={"help": "whether to resume training from latest epoch"}
    )
    step_eval_interval: int = field(
        default=-1, metadata={"help": "evaluation interval steps. `-1` if not use"}
    )
    epoch_eval_interval: int = field(
        default=1, metadata={"help": "evaluation interval epochs. `-1` if not use"}
    )
    eval_on_data: list = field(
        default_factory=lambda: ["dev"],
        metadata={"help": "evaluation on which data set(s)"},
    )
    select_best_on_data: str = field(
        default="dev", metadata={"help": "use which data to select best model"}
    )
    select_best_by_key: str = field(
        default="metric",
        metadata={"help": "the best model selection strategy, `metric` or `loss`"},
    )
    best_metric_field: str = field(
        default="micro.f1",
        metadata={
            "help": "metric value index to a dict (use `.` for solving nested dict)"
        },
    )
    save_every_ckpt: bool = field(
        default=False,
        metadata={"help": "whether to save every checkpoint on evaluation interval"},
    )
    save_best_ckpt: bool = field(
        default=True, metadata={"help": "whether to save the best checkpoint"}
    )
    final_eval_on_test: bool = field(
        default=True, metadata={"help": "whether to evaluate on final test set"}
    )
    # misc
    only_master_logging: bool = field(
        default=True, metadata={"help": "whether to logging in only master node"}
    )


if __name__ == "__main__":
    config = DefaultBaseConfig()
    print(config)
    print(asdict(config))



DEFAULT_PRF1_RESULT_DICT = {"p": 0.0, "r": 0.0, "f1": 0.0, "tp": 0, "fp": 0, "fn": 0}


def safe_division(numerator, denominator):
    try:
        val = numerator / denominator
    except ZeroDivisionError:
        val = 0.0
    return val


def calc_p_r_f1_from_tp_fp_fn(tp, fp, fn):
    p = safe_division(tp, tp + fp)
    r = safe_division(tp, tp + fn)
    f1 = safe_division(2 * p * r, p + r)

    return {"p": p, "r": r, "f1": f1, "tp": tp, "fp": fp, "fn": fn}



from torch.optim import Adam

from rex.data.collate_fn import re_collate_fn
from rex.data.data_manager import DataManager
from rex.data.dataset import CachedDataset
from rex.data.transforms.sent_re import CachedMCMLSentRETransform
from rex.metrics.classification import mcml_prf1
from rex.models.sent_pcnn import SentPCNN
from rex.tasks.simple_task import SimpleTask
from rex.utils.dict import PrettyPrintDict
from rex.utils.io import load_jsonlines
from rex.utils.registry import register
from rex.utils.tensor_move import detach_cpu_list


@register("task")
class MCMLSentRelationClassificationTask(SimpleTask):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init_transform(self):
        return CachedMCMLSentRETransform(
            self.config.max_seq_len,
            self.config.rel2id_filepath,
            self.config.emb_filepath,
        )

    def init_data_manager(self):
        dm = DataManager(
            self.config.train_filepath,
            self.config.dev_filepath,
            self.config.test_filepath,
            CachedDataset,
            self.transform,
            load_jsonlines,
            self.config.train_batch_size,
            self.config.eval_batch_size,
            re_collate_fn,
            use_stream_transform=False,
            debug_mode=self.config.debug_mode,
            dump_cache_dir=self.config.data_cache_dir,
        )
        return dm

    def init_model(self):
        m = SentPCNN(
            self.transform.label_encoder.num_tags,
            self.transform.vocab.size,
            self.config.dim_token_emb,
            self.config.max_seq_len,
            self.config.dim_pos_emb,
            self.config.num_filters,
            self.config.kernel_size,
            self.config.dropout,
        )
        m.token_embedding.from_pretrained(self.config.emb_filepath)
        return m

    def init_optimizer(self):
        return Adam(self.model.parameters(), lr=self.config.learning_rate)

    def get_eval_results(self, input_batches: list, output_batches: list) -> dict:
        preds = []
        golds = []
        for batch_input, batch_output in zip(input_batches, output_batches):
            batch_gold = detach_cpu_list(batch_input["labels"])
            batch_pred = detach_cpu_list(
                batch_output["pred"].ge(self.config.pred_threshold).long()
            )
            golds.extend(batch_gold)
            preds.extend(batch_pred)
        measures = mcml_prf1(preds, golds)
        return PrettyPrintDict(measures)

    def predict_api(self, text, head, tail):
        self.model.eval()
        batch = self.transform.predict_transform(
            {"text": text, "head": head, "tail": tail}
        )
        tensor_batch = self.data_manager.collate_fn([batch])
        if "labels" in tensor_batch:
            del tensor_batch["labels"]
        if "id" in tensor_batch:
            del tensor_batch["id"]
        result = self.model(**tensor_batch)
        outs = result["pred"]
        outs = detach_cpu_list(outs.ge(self.config.pred_threshold).long())[0]
        preds = []
        for type_idx, pred in enumerate(outs):
            if pred == 1:
                preds.append(self.transform.label_encoder.id2label[type_idx])

        return preds



from collections import defaultdict

from rex.metrics import (
    DEFAULT_PRF1_RESULT_DICT,
    calc_p_r_f1_from_tp_fp_fn,
    safe_division,
)


def get_measures_from_sets(gold_set: set, pred_set: set) -> dict:
    intersection = gold_set & pred_set
    tp = len(intersection)
    fp = len(pred_set - intersection)
    fn = len(gold_set - intersection)
    return calc_p_r_f1_from_tp_fp_fn(tp, fp, fn)


def tagging_prf1(gold_ents, pred_ents, type_idx=1) -> dict:
    """Get tagging P, R and F1 results

    Args:
        gold_ents: list of gold entities for each instance.
            Each entity **must** follow the ``(name, type, pos)`` format.
            e.g. [[(Alan, PER, (0, 4)), (Jack, PER, (5, 9))], [...], ...]
        pred_ents: ditto
        type_idx: index to entity type in entity tuples.
            e.g. `type_idx` in `(Alan, PER, (0, 4))` is `1`

    Returns:
        {
            "micro": {"p": 0.0, ...},
            "macro": {"p": 0.0, ...},
            "types": {
                "type1": {"p": 0.0, ...}
            }
        }
    """
    measure_results = {
        "micro": DEFAULT_PRF1_RESULT_DICT.copy(),
        "macro": DEFAULT_PRF1_RESULT_DICT.copy(),
        "types": defaultdict(lambda: DEFAULT_PRF1_RESULT_DICT.copy()),
    }
    for one_ins_gold_ents, one_ins_pred_ents in zip(gold_ents, pred_ents):
        _result = get_measures_from_sets(set(one_ins_gold_ents), set(one_ins_pred_ents))
        measure_results["micro"]["tp"] += _result["tp"]
        measure_results["micro"]["fp"] += _result["fp"]
        measure_results["micro"]["fn"] += _result["fn"]

        _type2ent = defaultdict(lambda: {"gold": set(), "pred": set()})
        for ent in one_ins_gold_ents:
            ent_type = ent[type_idx]
            _type2ent[ent_type]["gold"].add(ent)
        for ent in one_ins_pred_ents:
            ent_type = ent[type_idx]
            _type2ent[ent_type]["pred"].add(ent)
        for ent_type in _type2ent:
            _result = get_measures_from_sets(
                _type2ent[ent_type]["gold"], _type2ent[ent_type]["pred"]
            )
            measure_results["types"][ent_type]["tp"] += _result["tp"]
            measure_results["types"][ent_type]["fp"] += _result["fp"]
            measure_results["types"][ent_type]["fn"] += _result["fn"]

    # micro
    measure_results["micro"] = calc_p_r_f1_from_tp_fp_fn(
        measure_results["micro"]["tp"],
        measure_results["micro"]["fp"],
        measure_results["micro"]["fn"],
    )

    # for each type
    for ent_type in measure_results["types"]:
        measure_results["types"][ent_type] = calc_p_r_f1_from_tp_fp_fn(
            measure_results["types"][ent_type]["tp"],
            measure_results["types"][ent_type]["fp"],
            measure_results["types"][ent_type]["fn"],
        )

    measure_results["types"] = dict(measure_results["types"])

    # macro
    macro_results = defaultdict(list)
    for ent_type in measure_results["types"]:
        for key in measure_results["types"][ent_type]:
            macro_results[key].append(measure_results["types"][ent_type][key])
    for key in macro_results:
        measure_results["macro"][key] = safe_division(
            sum(macro_results[key]), len(macro_results[key])
        )

    return measure_results



import os

from datasets import load_dataset

from rex.data.label_encoder import LabelEncoder
from rex.utils.build_emb import build_emb
from rex.utils.io import dump_iterable, dump_jsonlines

cache_dir = "./data/cache"
dump_dir = "./data/formatted"
# downloaded emb file from https://github.com/Embedding/Chinese-Word-Vectors
input_embedding_filepath = (
    "/data4/tzhu/CCKS2020-FinEBody-nonNaN/data/sgns.sogounews.bigram-char"
)
output_embedding_filepath = "./data/formatted/vocab.emb"

datasets = load_dataset("peoples_daily_ner", cache_dir=cache_dir)
lbe = LabelEncoder()

if not os.path.exists(dump_dir):
    os.mkdir(dump_dir)


def extract(dataset):
    to_tag = dataset.info.features["ner_tags"].feature.int2str
    data = []
    tokens = set()
    for d in dataset:
        if len(d["tokens"]) < 1:
            continue
        tokens.update(d["tokens"])
        d["ner_tags"] = list(map(to_tag, d["ner_tags"]))
        lbe.update(d["ner_tags"])
        data.append(d)
    return data, tokens


vocab_files = []
for key in datasets:
    data, tokens = extract(datasets[key])
    dump_jsonlines(data, os.path.join(dump_dir, f"{key}.jsonl"))
    vocab_filepath = os.path.join(cache_dir, f"{key}.vocab")
    vocab_files.append(vocab_filepath)
    dump_iterable(tokens, vocab_filepath)

lbe.save_pretrained(os.path.join(dump_dir, "label2id.json"))
build_emb(input_embedding_filepath, output_embedding_filepath, *vocab_files)



import sys

from rex.utils.config import ConfigArgument, ConfigParser
from rex.utils.logging import logger
from rex.utils.registry import NAMESPACE_REGISTRY, call_register, get_registered

from .emb import emb
from .new import new
from .train import train


def main():
    args = ConfigParser.parse_cmd(
        ConfigArgument(
            "command",
            choices=list(NAMESPACE_REGISTRY["rex_init_call"].keys()),
            help="REx mode.",
        ),
        init_priority_args=False,
        cmd_args=sys.argv[1:2],
    )
    logger.info(f"Entering {args.command} mode...")

    if args.command in NAMESPACE_REGISTRY["rex_init_call"]:
        command = get_registered("rex_init_call", args.command)
        command(cmd_args=sys.argv[2:])
    else:
        print(f"Unknown command: {args.command}")
        sys.exit(1)


if __name__ == "__main__":
    main()



from typing import Iterable, Iterator


def flatten_all(iterable: Iterable) -> Iterator:
    for elem in iterable:
        if not isinstance(elem, list):
            yield elem
        else:
            yield from flatten_all(elem)



import torch
from omegaconf import OmegaConf
from torch.optim import Adam

from rex.data.collate_fn import subj_obj_span_collate_fn
from rex.data.data_manager import StreamTransformManager
from rex.data.dataset import StreamTransformDataset
from rex.data.transforms.entity_re import (
    StreamBERTSubjObjSpanTransform,
    StreamSubjObjSpanTransform,
)
from rex.metrics.triple import measure_triple
from rex.models.casrel import CasRel
from rex.tasks.base_task import TaskBase
from rex.utils.io import load_json, load_jsonlines
from rex.utils.logging import logger
from rex.utils.progress_bar import pbar
from rex.utils.tensor_move import move_to_cuda_device


class EntityRelationExtractionTask(TaskBase):
    def __init__(self, config: OmegaConf, **kwargs) -> None:
        super().__init__(config, **kwargs)

        rel2id = load_json(config.rel2id_filepath)
        self.transform = StreamBERTSubjObjSpanTransform(
            config.max_seq_len, rel2id, config.bert_model_dir
        )
        self.data_manager = StreamTransformManager(
            config.train_filepath,
            config.dev_filepath,
            config.test_filepath,
            StreamTransformDataset,
            self.transform,
            load_jsonlines,
            config.train_batch_size,
            config.eval_batch_size,
            subj_obj_span_collate_fn,
            debug_mode=config.debug_mode,
            load_train_data=config.load_train_data,
            load_dev_data=config.load_dev_data,
            load_test_data=config.load_test_data,
        )

        self.model = CasRel(
            config.bert_model_dir,
            self.transform.label_encoder.num_tags,
            pred_threshold=config.pred_threshold,
        )
        self.model.to(self.config.device)
        self.optimizer = Adam(self.model.parameters(), lr=config.learning_rate)

    def train(self):
        for epoch_idx in range(self.config.num_epochs):
            logger.info(f"Epoch: {epoch_idx}/{self.config.num_epochs}")
            self.model.train()
            loader = pbar(self.data_manager.train_loader, desc=f"Train(e{epoch_idx})")
            for batch in loader:
                batch = move_to_cuda_device(batch, self.config.device)
                used_keys = {
                    "token_ids",
                    "mask",
                    "subj_heads",
                    "subj_tails",
                    "subj_head",
                    "subj_tail",
                    "obj_head",
                    "obj_tail",
                }
                new_batch = {}
                for key, val in batch.items():
                    if key in used_keys:
                        new_batch[key] = val
                result = self.model(**new_batch)
                loss = result["loss"]
                loader.set_postfix({"loss": loss.item()})
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
            logger.info(loader)

            dev_measures = self.eval("dev")
            self.history["dev"].append(dev_measures)
            test_measures = self.eval("test")
            self.history["test"].append(test_measures)

            measures = dev_measures
            is_best = False
            if measures["f1"] > self.best_metric:
                is_best = True
                self.best_metric = measures["f1"]
                self.best_epoch = epoch_idx
                self.no_climbing_cnt = 0
            else:
                self.no_climbing_cnt += 1

            if is_best and self.config.save_best_ckpt:
                self.save_ckpt(epoch=epoch_idx)

            logger.info(
                f"Epoch: {epoch_idx}, is_best: {is_best}, Dev: {measures}, Test: {test_measures}"
            )

            if (
                self.config.num_early_stop > 0
                and self.no_climbing_cnt > self.config.num_early_stop
            ):
                break

        logger.info(
            (
                f"Trial finished. Best Epoch: {self.best_epoch}, Dev: {self.history['dev'][self.best_epoch]}, "
                f"Best Test: {self.history['test'][self.best_epoch]}"
            )
        )

    @torch.no_grad()
    def eval(self, dataset_name):
        self.model.eval()
        name2loader = {
            "train": self.data_manager.train_eval_loader,
            "dev": self.data_manager.dev_loader,
            "test": self.data_manager.test_loader,
        }
        loader = pbar(name2loader[dataset_name], desc=f"{dataset_name} Eval")
        preds = []
        golds = []
        for batch in loader:
            batch = move_to_cuda_device(batch, self.config.device)
            golds.extend(batch["triples"])
            used_keys = {"token_ids", "mask"}
            new_batch = {}
            for key, val in batch.items():
                if key in used_keys:
                    new_batch[key] = val
            result = self.model(**new_batch)
            pred_triples = result["pred"]
            preds.extend(pred_triples)
        logger.info(loader)
        measures = measure_triple(preds, golds)
        return measures["triple"]

    @torch.no_grad()
    def predict(self, text):
        self.model.eval()
        batch = self.transform.predict_transform(
            {
                "text": text,
            }
        )
        tensor_batch = self.data_manager.collate_fn([batch])
        tensor_batch = move_to_cuda_device(tensor_batch, self.config.device)
        result = self.model(
            token_ids=tensor_batch["token_ids"], mask=tensor_batch["mask"]
        )
        pred_triples = result["pred"][0]
        preds = []
        for triple in pred_triples:
            head_ent = batch["token_ids"][triple[0][0] : triple[0][1]]
            head_ent = self.transform.vocab.convert_ids_to_tokens(head_ent)
            head_ent = " ".join(head_ent)
            tail_ent = batch["token_ids"][triple[2][0] : triple[2][1]]
            tail_ent = self.transform.vocab.convert_ids_to_tokens(tail_ent)
            tail_ent = " ".join(tail_ent)
            relation = self.transform.label_encoder.id2label[triple[1]]
            preds.append((head_ent, relation, tail_ent))

        return preds



from tqdm import tqdm as _tqdm


class pbar(_tqdm):
    def __init__(self, *args, **kwargs):
        if "ncols" not in kwargs:
            kwargs["ncols"] = 80
        if "ascii" not in kwargs:
            kwargs["ascii"] = True
        super().__init__(*args, **kwargs)


class rbar(pbar):
    def __init__(self, iterator, *args, **kwargs):
        iterator = range(iterator, *args)
        super().__init__(iterator, **kwargs)



from rex.utils.build_emb import build_emb
from rex.utils.config import ConfigArgument, ConfigParser
from rex.utils.registry import register


@register("rex_init_call")
def emb(cmd_args=None):
    config = ConfigParser.parse_cmd(
        ConfigArgument("-r", "--raw-emb-filepath", help="raw embedding filepath"),
        ConfigArgument("-o", "--dump-emb-filepath", help="dumped embedding filepath"),
        ConfigArgument("-f", "--filepaths", nargs="*", help="raw text filepaths"),
        cmd_args=cmd_args,
    )
    build_emb(config.raw_emb_filepath, config.dump_emb_filepath, *config.filepaths)


if __name__ == "__main__":
    emb()



import random
from collections import defaultdict
from typing import Dict, Iterable, List, Optional

import numpy as np
from transformers.models.bert import BertTokenizerFast

from rex.data.label_encoder import LabelEncoder
from rex.data.transforms.base import TransformBase
from rex.data.vocab import Vocab


class StreamSubjObjSpanTransform(TransformBase):
    """
    Data transform for jointly entity-relation extraction task.
    """

    def __init__(
        self, max_seq_len: int, rel2id: Dict, vocab_filepath: str, pad="PAD", unk="UNK"
    ) -> None:
        super().__init__(max_seq_len)
        self.vocab = Vocab.from_pretrained(vocab_filepath, pad=pad, unk=unk)
        self.label_encoder = LabelEncoder(rel2id)

    def transform(self, data: Iterable) -> List[dict]:
        token_ids, mask = self.vocab.encode(
            data["tokens"], self.max_seq_len, update=False
        )

        subj_heads = np.zeros(self.max_seq_len, dtype=np.uint8)
        subj_tails = np.zeros(self.max_seq_len, dtype=np.uint8)
        one_subj_head = np.zeros(self.max_seq_len, dtype=np.uint8)
        one_subj_tail = np.zeros(self.max_seq_len, dtype=np.uint8)
        all_obj_head = np.zeros(
            (self.label_encoder.num_tags, self.max_seq_len), dtype=np.uint8
        )
        all_obj_tail = np.zeros(
            (self.label_encoder.num_tags, self.max_seq_len), dtype=np.uint8
        )
        one_obj_head = np.zeros(
            (self.label_encoder.num_tags, self.max_seq_len), dtype=np.uint8
        )
        one_obj_tail = np.zeros(
            (self.label_encoder.num_tags, self.max_seq_len), dtype=np.uint8
        )

        subj2objs = defaultdict(list)
        triples = []
        for relation_triple in data["relations"]:
            relation = self.label_encoder.encode([relation_triple[0]])[0]
            head_ent = data["entities"][relation_triple[1]]
            tail_ent = data["entities"][relation_triple[2]]
            triples.append(
                ((head_ent[1], head_ent[2]), relation, (tail_ent[1], tail_ent[2]))
            )
            if head_ent[2] <= self.max_seq_len:
                subj_heads[head_ent[1]] = 1
                subj_tails[head_ent[2] - 1] = 1
                subj2objs[(head_ent[1], head_ent[2])].append(
                    (relation, tail_ent[1], tail_ent[2])
                )
            if tail_ent[2] <= self.max_seq_len:
                all_obj_head[relation, tail_ent[1]] = 1
                all_obj_tail[relation, tail_ent[2] - 1] = 1

        one_subj = random.choice(list(subj2objs.keys()))
        one_subj_head[one_subj[0]] = 1
        one_subj_tail[one_subj[1] - 1] = 1
        for obj in subj2objs[one_subj]:
            one_obj_head[obj[0], obj[1]] = 1
            one_obj_tail[obj[0], obj[2] - 1] = 1

        formatted = {
            "token_ids": token_ids,
            "mask": mask,
            "subj_heads": subj_heads,
            "subj_tails": subj_tails,
            "one_subj": one_subj,
            "subj2objs": subj2objs,
            "triples": triples,
            "subj_head": one_subj_head,
            "subj_tail": one_subj_tail,
            "obj_head": one_obj_head.transpose(),
            "obj_tail": one_obj_tail.transpose(),
        }
        formatted.update(data)
        return formatted

    def predict_transform(self, obj: dict) -> dict:
        """Transform for prediction

        Args:
            obj: dict object with ``text`` field
        """
        obj["text"] = obj["text"].split()
        token_ids, mask = self.vocab.encode(
            list(obj["text"]), self.max_seq_len, update=False
        )
        d = {
            "token_ids": token_ids,
            "mask": mask,
        }
        return d


class StreamBERTSubjObjSpanTransform(TransformBase):
    """
    Data transform for jointly entity-relation extraction task, with BERT tokenizers.
    """

    def __init__(self, max_seq_len: int, rel2id: Dict, bert_model_dir: str) -> None:
        super().__init__(max_seq_len)
        self.vocab = None
        self.tokenizer = BertTokenizerFast.from_pretrained(bert_model_dir)
        self.label_encoder = LabelEncoder(rel2id)

    def tokenize(self, tokens: List[str]):
        token_list = []
        for token in tokens:
            new_token = self.tokenizer.tokenize(token)
            token_list.append(new_token)
        return token_list

    def encode(self, token_list: List[List], max_seq_len: int):
        final_tokens = [self.tokenizer.cls_token]
        final_mask = [1]
        flat_tokens = []
        for token in token_list:
            flat_tokens.extend(token)
        flat_tokens = flat_tokens[: max_seq_len - 2]
        final_tokens.extend(flat_tokens)
        final_tokens.append(self.tokenizer.sep_token)
        final_mask += [1] * len(flat_tokens)
        final_mask.append(1)
        assert len(final_tokens) == len(final_mask)
        assert len(final_mask) <= max_seq_len
        final_tokens += [self.tokenizer.pad_token] * (max_seq_len - len(final_tokens))
        final_tokens = self.tokenizer.convert_tokens_to_ids(final_tokens)
        final_mask += [0] * (max_seq_len - len(final_mask))

        return final_tokens, final_mask

    def get_offset_position(
        self, token_list: List[List], pos: int, offset: Optional[int] = 1
    ):
        """
        get new position after tokenized

        Args:
            offset: if [CLS] is used at the beginning of a sequence, offset must be 1
        """
        new_pos = -1
        record_idx = 0
        for idx, tokens in enumerate(token_list):
            if idx == pos:
                new_pos = record_idx
                break
            else:
                record_idx += len(tokens)
        return new_pos

    def transform(self, data: Iterable) -> List[dict]:
        token_list = self.tokenize(data["tokens"])
        token_ids, mask = self.encode(token_list, self.max_seq_len)

        subj_heads = np.zeros(self.max_seq_len, dtype=np.uint8)
        subj_tails = np.zeros(self.max_seq_len, dtype=np.uint8)
        one_subj_head = np.zeros(self.max_seq_len, dtype=np.uint8)
        one_subj_tail = np.zeros(self.max_seq_len, dtype=np.uint8)
        all_obj_head = np.zeros(
            (self.label_encoder.num_tags, self.max_seq_len), dtype=np.uint8
        )
        all_obj_tail = np.zeros(
            (self.label_encoder.num_tags, self.max_seq_len), dtype=np.uint8
        )
        one_obj_head = np.zeros(
            (self.label_encoder.num_tags, self.max_seq_len), dtype=np.uint8
        )
        one_obj_tail = np.zeros(
            (self.label_encoder.num_tags, self.max_seq_len), dtype=np.uint8
        )

        subj2objs = defaultdict(list)
        triples = []
        for relation_triple in data["relations"]:
            relation = self.label_encoder.encode([relation_triple[0]])[0]
            head_ent = data["entities"][relation_triple[1]]
            tail_ent = data["entities"][relation_triple[2]]
            head_ent_start = self.get_offset_position(token_list, head_ent[1])
            head_ent_end = self.get_offset_position(token_list, head_ent[2])
            tail_ent_start = self.get_offset_position(token_list, tail_ent[1])
            tail_ent_end = self.get_offset_position(token_list, tail_ent[2])
            if (
                0 < head_ent_start < self.max_seq_len
                and 0 < head_ent_end < self.max_seq_len
                and 0 < tail_ent_start < self.max_seq_len
                and 0 < tail_ent_end < self.max_seq_len
            ):
                triples.append(
                    (
                        (head_ent_start, head_ent_end),
                        relation,
                        (tail_ent_start, tail_ent_end),
                    )
                )
                subj_heads[head_ent_start] = 1
                subj_tails[head_ent_end - 1] = 1
                subj2objs[(head_ent_start, head_ent_end)].append(
                    (relation, tail_ent_start, tail_ent_end)
                )
                all_obj_head[relation, tail_ent_start] = 1
                all_obj_tail[relation, tail_ent_end - 1] = 1

        one_subj = None
        if len(subj2objs) > 0:
            one_subj = random.choice(list(subj2objs.keys()))
            one_subj_head[one_subj[0]] = 1
            one_subj_tail[one_subj[1] - 1] = 1
            for obj in subj2objs[one_subj]:
                one_obj_head[obj[0], obj[1]] = 1
                one_obj_tail[obj[0], obj[2] - 1] = 1

        formatted = {
            "token_ids": token_ids,
            "mask": mask,
            "subj_heads": subj_heads,
            "subj_tails": subj_tails,
            "one_subj": one_subj,
            "subj2objs": subj2objs,
            "triples": triples,
            "subj_head": one_subj_head,
            "subj_tail": one_subj_tail,
            "obj_head": one_obj_head.transpose(),
            "obj_tail": one_obj_tail.transpose(),
        }
        formatted.update(data)
        return formatted

    def predict_transform(self, obj: dict):
        """Transformation for prediction

        Args:
            obj: dict object with space tokenized ``text`` string in the field
        """
        token_list = self.tokenize(obj["text"].split())
        token_ids, mask = self.encode(token_list, max_seq_len=self.max_seq_len)
        d = {
            "token_ids": token_ids,
            "mask": mask,
        }
        return d



import os
import shutil
from pathlib import Path

import rex
from rex.utils.config import ConfigArgument, ConfigParser
from rex.utils.registry import register

REX_DIR = Path(rex.__file__).parent
TEMPLATES_DIR = REX_DIR.joinpath("templates")


@register("rex_init_call")
def new(cmd_args=None):
    new_args = ConfigParser.parse_cmd(
        ConfigArgument(
            "task_name", help="Name of the task (used for creating new task folder)"
        ),
        init_priority_args=False,
        cmd_args=cmd_args,
    )
    task_dir = Path(new_args.task_name)
    task_dir.mkdir(parents=True, exist_ok=False)
    src_dir = TEMPLATES_DIR.joinpath(new_args.task_type)
    for filename in os.listdir(src_dir):
        src_fp = src_dir.joinpath(filename)
        dest_fp = task_dir.joinpath(filename)
        shutil.copy2(src_fp, dest_fp)
    print(f"New {new_args.task_type} task in {task_dir.absolute()}")


if __name__ == "__main__":
    new()



from rex.tasks.base_task import TaskBase
from rex.utils.config import ConfigArgument, ConfigParser
from rex.utils.logging import logger
from rex.utils.registry import get_registered, import_module_and_submodules, register


@register("rex_init_call")
def train(cmd_args=None):
    parser, args = ConfigParser.parse_cmd_args(
        ConfigArgument(
            "-m",
            "--include-package",
            action="append",
            type=str,
            default=[],
            help="packages to load",
        ),
        cmd_args=cmd_args,
    )
    for package_name in getattr(args, "include_package", []):
        import_module_and_submodules(package_name)
    config = ConfigParser.parse_args_config(parser, args)
    if "task_type" not in config:
        raise ValueError(
            (
                "`task_type` must exist in config! "
                "If you are using the default base config, `task_type` is not initialised in it."
            )
        )
    TASK_CLASS = get_registered("task", config.task_type)
    task: TaskBase = TASK_CLASS(config)
    logger.info("Task initialised, ready to start")
    task.train()


if __name__ == "__main__":
    train()



from typing import Any, Dict, List, Optional

import torch

from rex.utils.iteration import flatten_all


class GeneralCollateFn(object):
    """DataLoader collate function for general purpose
        If you want to transform data while collating,
        inherent this class and override the ``update_data`` method

    Args:
        key2type (~dict): For type mapping initialization. When collating,
            only values whose keys are in ``key2type`` are grouped together,
            the others will be dropped.
        guessing (~bool): Guess type for each field. We recommend you
            to write your own ``key2type`` mapping when instantiate
            this collate obj and set ``guessing`` to False in case of
            any problems.
    """

    DEFAULT_TYPE_MAP = {int: torch.long, float: torch.float, str: None}

    def __init__(
        self, key2type: Optional[dict] = {}, guessing: Optional[bool] = False
    ) -> None:
        self.key2type = key2type
        self.guessing = guessing

    def update_type_mapping(self, key2type: dict):
        for key, val_type in key2type.items():
            self.key2type[key] = val_type

    def guess_types(self, instance: dict, update: Optional[bool] = False) -> dict:
        key2type = {}
        for key, val in instance.items():
            val_type = type(val)
            if val_type in self.DEFAULT_TYPE_MAP:
                key2type[key] = self.DEFAULT_TYPE_MAP[val_type]
            elif val_type == list:
                item_types = set()
                for item in flatten_all(val):
                    item_types.add(type(item))
                if len(item_types) == 1:
                    key2type[key] = self.DEFAULT_TYPE_MAP.get(item_types.pop())
                else:
                    key2type[key] = None
            else:
                key2type[key] = None

        if update:
            self.update_type_mapping(key2type)
        return key2type

    def _validate_data(self, data: List[dict]):
        """Validate if all data have the same keys"""
        all_keys = set()
        for d in data:
            if len(all_keys) == 0:
                all_keys.update(d.keys())
            else:
                assert (
                    set(d.keys()) == all_keys
                ), "Data instances does not have the same keys!"

    def update_data(self, data: List[dict]) -> List[dict]:
        """For those who transform data while collating, override this function"""
        return data

    def __call__(self, data: List[Dict[str, Any]]) -> Dict[str, Any]:
        self._validate_data(data)
        if len(self.key2type) == 0 and self.guessing:
            self.guess_types(data[0], update=True)

        data = self.update_data(data)

        final_data = {key: [] for key in self.key2type}
        for d in data:
            for key in self.key2type:
                final_data[key].append(d[key])

        for key, val_type in self.key2type.items():
            if val_type is not None and all(val is not None for val in final_data[key]):
                final_data[key] = torch.tensor(final_data[key], dtype=val_type)

        return final_data


def re_collate_fn(data):
    final_data = {
        "id": [],
        "token_ids": [],
        "mask": [],
        "labels": [],
        "head_pos": [],
        "tail_pos": [],
    }
    for d in data:
        for key in final_data:
            final_data[key].append(d[key])

    final_data["token_ids"] = torch.tensor(final_data["token_ids"], dtype=torch.long)
    final_data["mask"] = torch.tensor(final_data["mask"], dtype=torch.long)
    if all(x is not None for x in final_data["labels"]):
        final_data["labels"] = torch.tensor(final_data["labels"], dtype=torch.long)
    else:
        final_data["labels"] = None
    final_data["head_pos"] = torch.tensor(final_data["head_pos"], dtype=torch.long)
    final_data["tail_pos"] = torch.tensor(final_data["tail_pos"], dtype=torch.long)

    return final_data


def bag_re_collate_fn(data):
    final_data = {
        "id": [],
        "token_ids": [],
        "mask": [],
        "labels": [],
        "head_pos": [],
        "tail_pos": [],
        "scopes": [],
    }
    scope = []
    scope_begin = 0
    labels = []
    # data is a batch, for every bag in the batch:
    for bag in data:
        scope_ = [scope_begin]
        # for every instance in the bag
        for ins in bag:
            for key in final_data:
                if key != "scopes":
                    final_data[key].append(ins[key])
            scope_begin += 1
        labels.append(ins["labels"])
        scope_.append(scope_begin)
        scope.append(scope_)
    final_data["scopes"] = scope

    final_data["token_ids"] = torch.tensor(final_data["token_ids"], dtype=torch.long)
    final_data["mask"] = torch.tensor(final_data["mask"], dtype=torch.long)
    if all(x is not None for x in final_data["labels"]):
        final_data["labels"] = torch.tensor(labels, dtype=torch.long)
    else:
        final_data["labels"] = None
    final_data["head_pos"] = torch.tensor(final_data["head_pos"], dtype=torch.long)
    final_data["tail_pos"] = torch.tensor(final_data["tail_pos"], dtype=torch.long)

    return final_data


def subj_obj_span_collate_fn(data):
    final_data = {
        "id": [],
        "tokens": [],
        "entities": [],
        "relations": [],
        "token_ids": [],
        "mask": [],
        "subj_heads": [],
        "subj_tails": [],
        "one_subj": [],
        "subj2objs": [],
        "triples": [],
        "subj_head": [],
        "subj_tail": [],
        "obj_head": [],
        "obj_tail": [],
    }
    for d in data:
        for key in final_data:
            if key in d:
                final_data[key].append(d[key])

    final_data["token_ids"] = torch.tensor(final_data["token_ids"], dtype=torch.long)
    final_data["mask"] = torch.tensor(final_data["mask"], dtype=torch.long)
    for key in [
        "subj_heads",
        "subj_tails",
        "subj_head",
        "subj_tail",
        "obj_head",
        "obj_tail",
    ]:
        if final_data[key]:
            final_data[key] = torch.tensor(final_data[key], dtype=torch.float)

    return final_data



import logging
import sys
from typing import Optional

from loguru import logger
from tqdm import tqdm as _tqdm

logger.remove()
logger.add(lambda msg: _tqdm.write(msg, end=""))


class Logger(object):
    def __init__(
        self,
        name: str,
        log_path: Optional[str] = None,
        level: Optional[int] = logging.INFO,
    ):
        self.name = name
        self.log_path = log_path

        self.logger = logging.getLogger(self.name)
        self.logger.setLevel(level)

        fmt = "[%(asctime)-15s]-%(levelname)s-%(filename)s-%(lineno)d-%(process)d: %(message)s"
        datefmt = "%a %d %b %Y %H:%M:%S"
        formatter = logging.Formatter(fmt, datefmt)

        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(formatter)
        if not self.logger.handlers and log_path is not None:
            fh = logging.FileHandler(log_path)
            fh.setLevel(logging.INFO)
            fh.setFormatter(formatter)

            self.logger.addHandler(fh)
            self.logger.addHandler(stream_handler)

        self.debug = self.logger.debug
        self.info = self.logger.info
        self.warning = self.logger.warning
        self.error = self.logger.error

        self.logger.propagate = False
        sys.excepthook = self.handle_exception

    def handle_exception(self, exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return
        self.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))



from typing import Any, List, Type, Union


def find_all_positions(
    long: Union[List[Any], str], sub: Union[List[Any], str]
) -> List[List[int]]:
    """Find all sub list positions in the long list

    Args:
        long: list or string to positions in
        sub: short list or string to find

    Returns:
        all the appeared positions

    Raises:
        ValueError: if the len of `sub` is longer than the `long`
        ValueError: if the two input types are not the same

    Examples:
        >>> long_str = "123123123"
        >>> short_str = "123"
        >>> find_all_positions(long_str, short_str)
        [[0, 3], [3, 6], [6, 9]]
        >>> long_list = ["123", "1234", "12345"]
        >>> short_list = ["123"]
        >>> find_all_positions(long_list, short_list)
        [[0, 1]]
    """
    if isinstance(long, str) and isinstance(sub, str):
        long = list(long)
        sub = list(sub)
    if isinstance(long, list) and isinstance(sub, list):
        if len(sub) > len(long):
            raise ValueError("sub length is longer than the long")
        positions = []
        len_sub = len(sub)
        for idx in range(0, len(long) - len(sub) + 1):
            if sub == long[idx : idx + len_sub]:
                positions.append([idx, idx + len_sub])
        return positions
    else:
        raise TypeError("types of the two input must be str or list")


def construct_relative_positions(pos: int, max_length: int) -> List[int]:
    """Construct relative positions to a specified pos

    Args:
        pos: the pos that will be `0`
        max_length: max sequence length

    Returns:
        a list of relative positions

    Raises:
        ValueError: if pos is less than 0 or greater equal than max_length
    """
    if pos < 0 or pos >= max_length:
        raise ValueError(f"pos: {pos} is not in [0, {max_length})")
    positions = list(range(0, max_length, 1))
    positions = list(map(lambda x: abs(x - pos), positions))
    return positions



import torch.nn as nn


class SharedDropout(nn.Module):
    r"""
    SharedDropout differs from the vanilla dropout strategy in that
    the dropout mask is shared across one dimension.

    Args:
        p (float):
            The probability of an element to be zeroed. Default: 0.5.
        batch_first (bool):
            If ``True``, the input and output tensors are provided as ``[batch_size, seq_len, *]``.
            Default: ``True``.

    Examples:
        >>> x = torch.ones(1, 3, 5)
        >>> nn.Dropout()(x)
        tensor([[[0., 2., 2., 0., 0.],
                 [2., 2., 0., 2., 2.],
                 [2., 2., 2., 2., 0.]]])
        >>> SharedDropout()(x)
        tensor([[[2., 0., 2., 0., 2.],
                 [2., 0., 2., 0., 2.],
                 [2., 0., 2., 0., 2.]]])

    Reference:
        - https://github.com/yzhangcs/parser/blob/main/supar/modules/dropout.py
    """

    def __init__(self, p=0.5, batch_first=True):
        super().__init__()

        self.p = p
        self.batch_first = batch_first

    def __repr__(self):
        s = f"p={self.p}"
        if self.batch_first:
            s += f", batch_first={self.batch_first}"

        return f"{self.__class__.__name__}({s})"

    def forward(self, x):
        r"""
        Args:
            x (~torch.Tensor):
                A tensor of any shape.
        Returns:
            The returned tensor is of the same shape as `x`.
        """

        if self.training:
            if self.batch_first:
                mask = self.get_mask(x[:, 0], self.p).unsqueeze(1)
            else:
                mask = self.get_mask(x[0], self.p)
            x = x * mask

        return x

    @staticmethod
    def get_mask(x, p):
        return x.new_empty(x.shape).bernoulli_(1 - p) / (1 - p)



from typing import Any, Iterable


class TransformBase(object):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

    def transform(self, lines: Iterable):
        raise NotImplementedError

    def predict_transform(self, *args, **kwargs):
        raise NotImplementedError

    def __call__(self, *args: Any, **kwargs: Any):
        return self.transform(*args, **kwargs)



import re
from typing import List, Optional


def sent_seg(
    text: str,
    special_seg_indicators: Optional[List] = None,
    lang: Optional[str] = "zh",
    punctuations: Optional[set] = None,
    quotation_seg_mode: Optional[bool] = True,
) -> List[str]:
    """Cut texts into sentences (in Chinese or English).

    Args:
        text: texts ready to be cut
        special_seg_indicators: some special segment indicators and
            their replacement ( [indicator, replacement] ), in baike data,
            this argument could be ``[('###', '\\n'), ('%%%', ' '), ('%%', ' ')]``
        lang: languages that your corpus is, support ``zh`` for Chinese
            and ``en`` for English now.
        punctuations: you can split the texts by specified punctuations.
            texts will not be splited by ``;``, so you can specify them by your own.
        quotation_seg_mode: if True, the quotations will be regarded as a
            part of the former sentence.
            e.g. ``我说：“翠花，上酸菜。”她说：“欸，好嘞。”``
            the text will be splited into
            ``['我说：“翠花，上酸菜。”', '她说：“欸，好嘞。”']``, other than
            ``['我说：“翠花，上酸菜。', '”她说：“欸，好嘞。”']``

    Returns:
        a list of strings, which are splited sentences.

    Raises:
        ValueError: if text is not string
    """
    # if texts are not in string format, raise an error
    if not isinstance(text, str):
        raise ValueError

    # if the text is empty, return a list with an empty string
    if len(text) == 0:
        return []

    text_return = text

    # segment on specified indicators
    # special indicators standard, like [('###', '\n'), ('%%%', '\t'), ('\s', '')]
    if special_seg_indicators:
        for indicator in special_seg_indicators:
            text_return = re.sub(indicator[0], indicator[1], text_return)

    if lang == "zh":
        punkt = {"。", "？", "！", "…"}
    elif lang == "en":
        punkt = {".", "?", "!"}
    if punctuations:
        punkt = punkt | punctuations

    if quotation_seg_mode:
        text_return = re.sub(
            "([%s]+[’”`'\"]*)" % ("".join(punkt)), "\\1\n", text_return
        )
    else:
        text_return = re.sub("([{}])".format("".join(punkt)), "\\1\n", text_return)

    # drop sentences with no length
    return [
        s.strip()
        for s in filter(
            lambda x: len(x.strip()) == 1
            and x.strip() not in punkt
            or len(x.strip()) > 0,
            text_return.split("\n"),
        )
    ]



from typing import Iterable, Optional

import torch
import torch.nn as nn

from rex.modules.dropout import SharedDropout


class FFN(nn.Module):
    """
    Multi-layer feed-forward neural networks

    Args:
        input_dim: input dimension
        output_dim: output dimension
        mid_dims: middle dimensions, if None, FFN is equals to `nn.Linear` with dropout
        dropout: dropout rate
        act_fn: activation function (module class without instantiated)

    Input:
        hidden: hidden states
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        mid_dims: Optional[Iterable[int]] = None,
        dropout: Optional[float] = 0.5,
        act_fn: Optional[nn.Module] = nn.ReLU(),
    ):
        super().__init__()

        if mid_dims is None:
            self.ffn = nn.Sequential(
                nn.Linear(input_dim, output_dim), nn.Dropout(dropout)
            )
        else:
            mid_dims = list(mid_dims)
            mid_dims.insert(0, input_dim)
            mid_dims.append(output_dim)
            len_mid_dims = len(mid_dims)
            modules = []
            for i in range(len_mid_dims - 2):
                modules.extend(
                    [
                        nn.Linear(mid_dims[i], mid_dims[i + 1]),
                        nn.Dropout(dropout),
                        act_fn,
                    ]
                )
            modules.append(nn.Linear(mid_dims[-2], mid_dims[-1]))
            self.ffn = nn.Sequential(*modules)

    def forward(self, hidden):
        return self.ffn(hidden)


class MLP(nn.Module):
    """Implements Multi-layer Perception."""

    def __init__(
        self,
        input_size,
        output_size,
        mid_size=None,
        num_mid_layer=1,
        act_fn=torch.relu,
        dropout=0.1,
    ):
        super(MLP, self).__init__()

        assert num_mid_layer >= 1
        if mid_size is None:
            mid_size = input_size

        self.act_fn = act_fn
        self.input_fc = nn.Linear(input_size, mid_size)
        self.out_fc = nn.Linear(mid_size, output_size)
        if num_mid_layer > 1:
            self.mid_fcs = nn.ModuleList(
                nn.Linear(mid_size, mid_size) for _ in range(num_mid_layer - 1)
            )
        else:
            self.mid_fcs = []
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        x = self.dropout(self.act_fn(self.input_fc(x)))
        for mid_fc in self.mid_fcs:
            x = self.dropout(self.act_fn(mid_fc(x)))
        x = self.out_fc(x)
        return x


class SharedDropoutMLP(nn.Module):
    r"""
    Applies a linear transformation together with a non-linear activation to the incoming tensor:
    :math:`y = \mathrm{Activation}(x A^T + b)`
    Args:
        n_in (~torch.Tensor):
            The size of each input feature.
        n_out (~torch.Tensor):
            The size of each output feature.
        dropout (float):
            If non-zero, introduce a :class:`SharedDropout` layer on the output with this dropout ratio. Default: 0.
        activation (bool):
            Whether to use activations. Default: True.
    """

    def __init__(self, n_in, n_out, dropout=0, activation=True):
        super().__init__()

        self.n_in = n_in
        self.n_out = n_out
        self.linear = nn.Linear(n_in, n_out)
        self.activation = (
            nn.LeakyReLU(negative_slope=0.1) if activation else nn.Identity()
        )
        self.dropout = SharedDropout(p=dropout)

        self.reset_parameters()

    def __repr__(self):
        s = f"n_in={self.n_in}, n_out={self.n_out}"
        if self.dropout.p > 0:
            s += f", dropout={self.dropout.p}"

        return f"{self.__class__.__name__}({s})"

    def reset_parameters(self):
        nn.init.orthogonal_(self.linear.weight)
        nn.init.zeros_(self.linear.bias)

    def forward(self, x):
        r"""
        Args:
            x (~torch.Tensor):
                The size of each input feature is `n_in`.
        Returns:
            A tensor with the size of each output feature `n_out`.
        """

        x = self.linear(x)
        x = self.activation(x)
        x = self.dropout(x)

        return x



import torch.nn as nn


class ModelBase(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()

    def forward(self, *args, **kwargs) -> dict:
        """Model forwarding
        To override this function, you must make sure `**kwargs` is in your args

        Returns:
            return a dict of results, loss and preds must be included
            example: {"loss": loss, "preds": preds}
        """
        raise NotImplementedError



from collections import defaultdict
from typing import Iterable, Mapping, Optional

import numpy as np
from sklearn import metrics

from rex.metrics import DEFAULT_PRF1_RESULT_DICT, calc_p_r_f1_from_tp_fp_fn


def accuracy(preds, golds):
    if len(preds) == 0 or len(golds) == 0:
        raise ValueError("Preds or golds is empty.")
    correct = total = 0
    for pred, gold in zip(preds, golds):
        total += 1
        if pred == gold:
            correct += 1
    return correct / total


def mcml_prf1(preds, golds):
    if len(preds) == 0 or len(golds) == 0:
        raise ValueError("Preds or golds is empty.")
    measure_results = defaultdict(lambda: {"p": 0.0, "r": 0.0, "f1": 0.0})
    supports = defaultdict(lambda: {"tp": 0.0, "fp": 0.0, "fn": 0.0})
    for pred, gold in zip(preds, golds):
        if len(pred) != len(gold):
            raise ValueError(
                f"Pred: {pred} cannot be matched with gold: {gold} in length."
            )
        for type_idx, (p, g) in enumerate(zip(pred, gold)):
            if p == 1 and g == 1:
                supports[type_idx]["tp"] += 1
            elif p == 1 and g == 0:
                supports[type_idx]["fp"] += 1
            elif p == 0 and g == 1:
                supports[type_idx]["fn"] += 1
    g_tp = g_fp = g_fn = 0.0
    ps = []
    rs = []
    f1s = []
    for type_idx in supports:
        measure_results[type_idx] = calc_p_r_f1_from_tp_fp_fn(
            supports[type_idx]["tp"], supports[type_idx]["fp"], supports[type_idx]["fn"]
        )
        ps.append(measure_results[type_idx]["p"])
        rs.append(measure_results[type_idx]["r"])
        f1s.append(measure_results[type_idx]["f1"])
        g_tp += supports[type_idx]["tp"]
        g_fp += supports[type_idx]["fp"]
        g_fn += supports[type_idx]["fn"]

    measure_results["micro"] = calc_p_r_f1_from_tp_fp_fn(g_tp, g_fp, g_fn)
    measure_results["macro"]["p"] = sum(ps) / len(ps)
    measure_results["macro"]["r"] = sum(rs) / len(rs)
    measure_results["macro"]["f1"] = sum(f1s) / len(f1s)

    return measure_results


def mc_prf1(
    preds,
    golds,
    num_classes=-1,
    ignore_labels: Optional[Iterable] = None,
    label_idx2name: Optional[Mapping[int, str]] = None,
):
    """
    get multi-class classification metrics

    Args:
        num_classes: guess the current number of classes if < 0
        ignore_labels: labels to ignore when calculating f1 scores,
            especially helpful when calculating non-NA metrics
    """
    if len(preds) == 0 or len(golds) == 0:
        raise ValueError("Preds or golds is empty.")

    measure_results = {
        "micro": DEFAULT_PRF1_RESULT_DICT.copy(),
        "macro": DEFAULT_PRF1_RESULT_DICT.copy(),
        "types": defaultdict(lambda: DEFAULT_PRF1_RESULT_DICT.copy()),
    }

    if num_classes > 0:
        # specified number of classes
        labels = list(range(num_classes))
    else:
        # guess the number of classes through current input
        labels = None

    mcm = metrics.multilabel_confusion_matrix(
        golds, preds, sample_weight=None, labels=labels, samplewise=False
    )

    tp_list, fp_list, fn_list = [], [], []
    if ignore_labels is None:
        ignore_labels = []
    for tmp_label in range(mcm.shape[0]):
        if tmp_label not in ignore_labels:
            tp_list.append(mcm[tmp_label, 1, 1])
            fp_list.append(mcm[tmp_label, 0, 1])
            fn_list.append(mcm[tmp_label, 1, 0])

    tp_list = np.stack(tp_list)
    fp_list = np.stack(fp_list)
    fn_list = np.stack(fn_list)

    # micro-averaged scores
    measure_results["micro"] = calc_p_r_f1_from_tp_fp_fn(
        tp_list.sum(), fp_list.sum(), fn_list.sum()
    )

    # macro-averaged scores
    results = calc_p_r_f1_from_tp_fp_fn(tp_list, fp_list, fn_list)

    measure_results["macro"]["p"] = results["p"].mean()
    measure_results["macro"]["r"] = results["r"].mean()
    measure_results["macro"]["f1"] = results["f1"].mean()
    measure_results["macro"]["tp"] = tp_list.mean()
    measure_results["macro"]["fp"] = fp_list.mean()
    measure_results["macro"]["fn"] = fn_list.mean()

    if label_idx2name is None:
        label_idx2name = {}
    for idx, p, r, f1, tp, fp, fn in zip(
        range(mcm.shape[0]),
        results["p"],
        results["r"],
        results["f1"],
        tp_list,
        fp_list,
        fn_list,
    ):
        if idx in label_idx2name:
            idx_name = label_idx2name.get(idx)
        else:
            idx_name = idx
        measure_results["types"][idx_name] = {
            "p": p,
            "r": r,
            "f1": f1,
            "tp": tp,
            "fp": fp,
            "fn": fn,
        }

    measure_results["types"] = dict(measure_results["types"])
    return measure_results



from collections import defaultdict
from typing import Iterable, List, Optional

from rex.data.label_encoder import LabelEncoder
from rex.data.transforms.base import TransformBase
from rex.data.vocab import Vocab
from rex.utils.logging import logger
from rex.utils.mask import construct_piecewise_mask
from rex.utils.position import construct_relative_positions, find_all_positions
from rex.utils.progress_bar import pbar


class CachedMCMLSentRETransform(TransformBase):
    """
    Data transform for cached multi-class multi-label sentence-level relation extraction task.
    """

    def __init__(self, max_seq_len, rel2id_filepath, emb_filepath) -> None:
        super().__init__()
        self.max_seq_len = max_seq_len
        self.vocab = Vocab.from_pretrained(
            emb_filepath,
            include_weights=True,
            init_pad_unk_emb=False,
            include_pad=False,
            include_unk=False,
        )
        self.label_encoder = LabelEncoder.from_pretrained(rel2id_filepath)

    def transform(
        self,
        data: Iterable,
        desc: Optional[str] = "Transform",
        debug: Optional[bool] = False,
    ) -> List[dict]:
        num_truncated_rels = 0
        final_data = []
        if debug:
            data = data[:5]
        transform_loader = pbar(data, desc=desc)

        for d in transform_loader:
            ent_validation = []
            for ent in d["entities"]:
                if ent[2] > self.max_seq_len:
                    ent_validation.append(False)
                    continue
                ent_validation.append(True)
            valid_ent_pair2rels = defaultdict(set)
            for rel in d["relations"]:
                if ent_validation[rel[1]] is False or ent_validation[rel[2]] is False:
                    num_truncated_rels += 1
                    continue
                valid_ent_pair2rels[(rel[1], rel[2])].add(
                    self.label_encoder.label2id[rel[0]]
                )
            if len(valid_ent_pair2rels) == 0:
                continue
            token_ids, _ = self.vocab.encode(
                d["tokens"], self.max_seq_len, update=False
            )
            for ent_pair, rels in valid_ent_pair2rels.items():
                head_pos = construct_relative_positions(
                    d["entities"][ent_pair[0]][1], self.max_seq_len
                )
                tail_pos = construct_relative_positions(
                    d["entities"][ent_pair[1]][1], self.max_seq_len
                )
                final_data.append(
                    {
                        "id": d["id"],
                        "token_ids": token_ids,
                        "mask": construct_piecewise_mask(
                            d["entities"][ent_pair[0]][1],
                            d["entities"][ent_pair[1]][1],
                            min(len(d["tokens"]), self.max_seq_len),
                            self.max_seq_len,
                        ),
                        "labels": rels,
                        "head_pos": head_pos,
                        "tail_pos": tail_pos,
                    }
                )
        final_data = list(filter(lambda x: x is not None, final_data))
        for d in final_data:
            d["labels"] = self.label_encoder.convert_to_multi_hot(d["labels"])
        logger.info(transform_loader)
        logger.warning(f"#truncated_rels: {num_truncated_rels}")
        return final_data

    def predict_transform(self, obj: dict):
        """Transform for prediction

        Args:
            obj: dict object with ``text`` (str), ``head`` (str) and ``tail`` (str) fields
        """
        if obj["head"] not in obj["text"] or obj["tail"] not in obj["text"]:
            raise ValueError(f"{obj['head']} or {obj['tail']} is not in {obj['text']}")
        head_pos = find_all_positions(obj["text"], obj["head"])[0]
        tail_pos = find_all_positions(obj["text"], obj["tail"])[0]
        if head_pos[1] > self.max_seq_len:
            logger.warning("head entity truncated")
            head_pos = [self.max_seq_len - 1, self.max_seq_len]
        if tail_pos[1] > self.max_seq_len:
            logger.warning("tail entity truncated")
            tail_pos = [self.max_seq_len - 1, self.max_seq_len]
        token_ids, _ = self.vocab.encode(
            list(obj["text"]), self.max_seq_len, update=False
        )
        d = {
            "id": "",
            "token_ids": token_ids,
            "mask": construct_piecewise_mask(
                head_pos[0],
                tail_pos[0],
                min(len(obj["text"]), self.max_seq_len),
                self.max_seq_len,
            ),
            "labels": None,
            "head_pos": construct_relative_positions(head_pos[0], self.max_seq_len),
            "tail_pos": construct_relative_positions(tail_pos[0], self.max_seq_len),
        }
        return d



# ref: https://github.com/bytedance/ParaGen
import importlib
import os
import pkgutil
import sys
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, Optional, Set

CLASS_REGISTRY = {}


def register_class(cls):
    """
    Register a class with its name
    Args:
        cls: a new class fro registration
    """
    name = cls.__name__
    # if name in CLASS_REGISTRY:
    #     raise ValueError(f"Cannot register duplicate `CLASS_REGISTRY` class ({name})")
    CLASS_REGISTRY[name] = cls
    return cls


def get_registered_class(name):
    # if name not in CLASS_REGISTRY:
    #     raise ValueError(f"{name} not registered!")
    return CLASS_REGISTRY[name]


NAMESPACE_REGISTRY = defaultdict(dict)


def register(namespace: str):
    def register_on_namespace(call: Callable):
        cname = call.__name__
        # if cname in NAMESPACE_REGISTRY[namespace]:
        #     raise ValueError(
        #         f"Cannot register duplicate {cname} in `NAMESPACE_REGISTRY[{namespace}]`"
        #     )
        NAMESPACE_REGISTRY[namespace][cname] = call
        return call

    return register_on_namespace


def get_registered(namespace: str, call_name: str):
    # if call_name not in NAMESPACE_REGISTRY[namespace]:
    #     raise ValueError(f"{call_name} not registered in {namespace}!")
    call = NAMESPACE_REGISTRY[namespace][call_name]
    return call


def _get_module_name(path: str):
    return ".".join(path.replace("/__init__.py", "").split("/")).replace(".py", "")


def _import_module(modules_dir: str) -> None:
    simplified_module_dir = modules_dir[modules_dir.find("rex/") :]
    for file in os.listdir(modules_dir):
        path = os.path.join(modules_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            module = ".".join(
                os.path.join(simplified_module_dir, file)
                .replace("/__init__.py", "")
                .split("/")
            ).replace(".py", "")
            importlib.import_module(module)


def call_register(path):
    """
    Args:
        path: __file__
    """
    module_dir = os.path.dirname(path)
    if os.path.exists(module_dir):
        _import_module(module_dir)
    else:
        raise ValueError(f"Dir {module_dir} not exists!")


@contextmanager
def push_python_path(path):
    """
    Prepends the given path to `sys.path`.
    This method is intended to use with `with`, so after its usage, its value willbe removed from
    `sys.path`.
    """
    # In some environments, such as TC, it fails when sys.path contains a relative path, such as ".".
    path = Path(path).resolve()
    path = str(path)
    sys.path.insert(0, path)
    try:
        yield
    finally:
        # Better to remove by value, in case `sys.path` was manipulated in between.
        sys.path.remove(path)


def import_module_and_submodules(
    package_name: str, exclude: Optional[Set[str]] = None
) -> None:
    """
    Import all public submodules under the given package.
    Primarily useful so that people using AllenNLP as a library
    can specify their own custom packages and have their custom
    classes get loaded and registered.
    Idea borrowed from [AllenNLP]_.

    References:
        .. [AllenNLP] https://github.com/allenai/allennlp/blob/068407e3d476750ee75fd52840be7a160b693760/allennlp/common/util.py#L266
    """
    # take care of None
    exclude = exclude if exclude else set()
    if package_name in exclude:
        return

    importlib.invalidate_caches()

    # For some reason, python doesn't always add this by default to your path, but you pretty much
    # always want it when using `--include-package`.  And if it's already there, adding it again at
    # the end won't hurt anything.
    with push_python_path("."):
        # Import at top level
        module = importlib.import_module(package_name)
        path = getattr(module, "__path__", [])
        path_string = "" if not path else path[0]

        # walk_packages only finds immediate children, so need to recurse.
        for module_finder, name, _ in pkgutil.walk_packages(path):
            # Sometimes when you import third-party libraries that are on your path,
            # `pkgutil.walk_packages` returns those too, so we need to skip them.
            if path_string and module_finder.path != path_string:  # type: ignore[union-attr]
                continue
            if name.startswith("_"):
                # skip directly importing private subpackages
                continue
            if name.startswith("test"):
                continue
            subpackage = f"{package_name}.{name}"
            import_module_and_submodules(subpackage, exclude=exclude)


def import_submodules(package_name: str) -> None:
    """
    Import all submodules under the given package.
    Primarily useful so that people using REx as a library
    can specify their own custom packages and have their custom
    classes get loaded and registered.
    Idea borrowed from [AllenNLP]_.

    References:
        .. [AllenNLP] https://github.com/allenai/allennlp/blob/068407e3d476750ee75fd52840be7a160b693760/allennlp/common/util.py#L266
    """
    importlib.invalidate_caches()

    # Import at top level
    module = importlib.import_module(package_name)
    path = getattr(module, "__path__", "")

    # walk_packages only finds immediate children, so need to recurse.
    for _, name, _ in pkgutil.walk_packages(path):
        subpackage = f"{package_name}.{name}"
        import_submodules(subpackage)



from collections import defaultdict

from rex.metrics import calc_p_r_f1_from_tp_fp_fn


def measure_triple(preds, golds):
    result = {
        "triple": {"p": 0.0, "r": 0.0, "f1": 0.0},
        "subject": {"p": 0.0, "r": 0.0, "f1": 0.0},
        "object": {"p": 0.0, "r": 0.0, "f1": 0.0},
        "relation": {"p": 0.0, "r": 0.0, "f1": 0.0},
    }
    middle_stat = defaultdict(lambda: {"tp": 0, "fp": 0, "fn": 0})
    for pred, gold in zip(preds, golds):
        pred_triples = set(pred)
        gold_triples = set(gold)
        middle_stat["triple"]["tp"] += len(pred_triples & gold_triples)
        middle_stat["triple"]["fp"] += len(pred_triples - gold_triples)
        middle_stat["triple"]["fn"] += len(gold_triples - pred_triples)

        pred_relations = set(x[1] for x in pred_triples)
        gold_relations = set(x[1] for x in gold_triples)
        middle_stat["relation"]["tp"] += len(pred_relations & gold_relations)
        middle_stat["relation"]["fp"] += len(pred_relations - gold_relations)
        middle_stat["relation"]["fn"] += len(gold_relations - pred_relations)

        pred_subjs = set(x[0] for x in pred_triples)
        gold_subjs = set(x[0] for x in gold_triples)
        middle_stat["subject"]["tp"] += len(pred_subjs & gold_subjs)
        middle_stat["subject"]["fp"] += len(pred_subjs - gold_subjs)
        middle_stat["subject"]["fn"] += len(gold_subjs - pred_subjs)

        pred_objs = set(x[2] for x in pred_triples)
        gold_objs = set(x[2] for x in gold_triples)
        middle_stat["object"]["tp"] += len(pred_objs & gold_objs)
        middle_stat["object"]["fp"] += len(pred_objs - gold_objs)
        middle_stat["object"]["fn"] += len(gold_objs - pred_objs)

    for measure in result:
        result[measure] = calc_p_r_f1_from_tp_fp_fn(**middle_stat[measure])
        result[measure].update(middle_stat[measure])

    return result



import datetime
import functools
from datetime import timedelta
from typing import Callable, Optional

from rex.utils.logging import logger


def safe_try(
    func_placeholder: Optional[Callable] = None,
    start_msg: Optional[str] = None,
    end_msg: Optional[str] = None,
):
    def func_wrapper(func):
        @functools.wraps(func)
        def try_func(*args, **kwargs):
            if start_msg is not None:
                logger.info(str(start_msg))

            err_happened = False
            start_dt = datetime.datetime.now()

            try:
                return func(*args, **kwargs)
            except KeyboardInterrupt:
                err_happened = True
                raise KeyboardInterrupt
            except Exception as err:
                err_happened = True
                logger.exception(err)
            finally:
                time_delta = datetime.datetime.now() - start_dt
                if not err_happened:
                    logger.debug(f"Func `{func.__name__}` finished without err.")
                if end_msg is not None:
                    logger.info(str(end_msg))
                logger.info(f"Func: `{func.__name__}` call time: {str(time_delta)}")

        return try_func

    if func_placeholder is None:
        return func_wrapper
    else:
        return func_wrapper(func_placeholder)



import math
from typing import Iterable, Optional

import torch
import torch.nn as nn


class MultiKernelCNN(nn.Module):
    """
    Multi-kernel Convolutional Neural Network Module

    Args:
        in_channel: dimension of input features
        num_filters: number of filters in CNN
        kernel_sizes: kernel sizes, e.g. ``[1, 3, 5]``
        dropout: dropout rate

    Returns:
        torch.Tensor: Shape is :math:`x \\in \\mathbb{R}^{B \\times N \\cdot F}`,
            where :math:`B` is ``batch_size``,
            :math:`N` means the length of ``kernel_sizes``
            and :math:`F` is the number of CNN filters
    """

    def __init__(
        self,
        in_channel: int,
        num_filters: int,
        kernel_sizes: Iterable,
        dropout: Optional[float] = 0.5,
    ):
        super().__init__()
        self.convs = nn.ModuleList(
            [
                nn.Conv1d(in_channel, num_filters, ks, padding=math.floor((ks - 1) / 2))
                for ks in kernel_sizes
            ]
        )
        self.dropout = nn.Dropout(dropout)

    def forward(self, hidden):
        batch_size = hidden.shape[0]
        # hidden: (batch_size, seq_len, in_channel) -> (batch_size, in_channel, seq_len)
        hidden = hidden.permute(0, 2, 1)
        # (len(kernel_sizes), batch_size, num_filters, seq_len)
        hidden = torch.stack([conv(hidden) for conv in self.convs], dim=0)
        # max pooling: (len(kernel_sizes), batch_size, num_filters)
        hidden = hidden.max(dim=-1)[0]
        # (batch_size, len(kernel_sizes), num_filters)
        hidden = hidden.permute(1, 0, 2)
        # (batch_size, len(kernel_sizes) * num_filters)
        hidden = hidden.reshape(batch_size, -1)
        hidden = self.dropout(hidden)
        return hidden



import logging
import os
from pathlib import Path
from typing import Optional

import torch
from accelerate import Accelerator
from omegaconf import OmegaConf
from torch import distributed
from torch.nn import parallel

from rex.tasks import (
    BEST_CHECKPOINT_FILENAME_TEMPLATE,
    BEST_IDENTIFIER,
    CHECKPOINT_DIRNAME,
    CHECKPOINT_FILENAME_TEMPLATE,
    CONFIG_PARAMS_FILENAME,
)
from rex.utils.initialization import init_all
from rex.utils.logging import logger


class TaskBase(object):
    def __init__(
        self,
        config: OmegaConf,
        initialize: Optional[bool] = True,
        makedirs: Optional[bool] = True,
        dump_configfile: Optional[bool] = True,
    ) -> None:
        self.config = config

        # `batch_size` in config is the total batch size number
        self.accelerator = Accelerator()
        self.device = self.accelerator.device

        self.model = None
        self.optimizer = None

        self.history = {
            "epoch": {
                "train": {"loss": {}, "metrics": {}},
                "train_eval": {"loss": {}, "metrics": {}},
                "dev": {"loss": {}, "metrics": {}},
                "test": {"loss": {}, "metrics": {}},
            },
            "step": {
                "train": {"loss": {}, "metrics": {}},
                "train_eval": {"loss": {}, "metrics": {}},
                "dev": {"loss": {}, "metrics": {}},
                "test": {"loss": {}, "metrics": {}},
            },
            "no_climbing_epoch_cnt": 0,
            "no_climbing_step_cnt": 0,
            "best_metric": -100.0,
            "best_loss": float("inf"),
            "best_epoch": -1,
            "best_step": -1,
            "curr_epoch": 0,
            "curr_batch": 0,
            "total_steps": 0,
            "current_train_loss": 0.0,
        }

        if initialize:
            init_all(
                config.task_dir, seed=config.random_seed, set_cudnn=True, config=None
            )

        config_string = OmegaConf.to_yaml(config, resolve=True)
        self.logging(config_string)

        if makedirs:
            if not os.path.exists(config.task_dir):
                os.makedirs(config.task_dir)
            else:
                self.logging(f"Overwrite task dir: {config.task_dir}", level="WARNING")

        if dump_configfile:
            OmegaConf.save(
                config,
                os.path.join(config.task_dir, CONFIG_PARAMS_FILENAME),
                resolve=True,
            )

    def train(self, *args, **kwargs):
        raise NotImplementedError

    def eval(self, *args, **kwargs):
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        raise NotImplementedError

    def reset_history(self, reset_all: Optional[bool] = False):
        self.history["no_climbing_epoch_cnt"] = 0
        self.history["no_climbing_step_cnt"] = 0
        self.history["best_metric"] = -100.0
        self.history["best_loss"] = float("inf")
        self.history["best_epoch"] = -1
        self.history["best_step"] = -1
        self.history["curr_epoch"] = 0
        self.history["curr_batch"] = 0
        self.history["total_steps"] = 0
        self.history["current_train_loss"] = 0.0

        if reset_all:
            for type_key in ["epoch", "step"]:
                for data_type in ["train", "train_eval", "dev", "test"]:
                    for measure_type in ["loss", "metrics"]:
                        self.history[type_key][data_type][measure_type].clear()

    def load(
        self,
        path: str,
        load_config: Optional[bool] = False,
        load_model: Optional[bool] = True,
        load_optimizer: Optional[bool] = False,
        load_history: Optional[bool] = True,
    ):
        self.accelerator.wait_for_everyone()
        if os.path.exists(path):
            self.logging("Resume checkpoint from {}".format(path))
        else:
            raise ValueError("Checkpoint does not exist, {}".format(path))

        if torch.cuda.device_count() == 0:
            self.logging("Load store_dict into cpu", level="DEBUG")
            store_dict = torch.load(path, map_location="cpu")
        else:
            self.logging(
                f"Load store_dict into {self.accelerator.device}", level="DEBUG"
            )
            store_dict = torch.load(path, map_location=self.accelerator.device)

        if load_config:
            self.config = OmegaConf.load(
                os.path.join(self.config.task_dir, CONFIG_PARAMS_FILENAME)
            )

        if load_model:
            if self.model and "model_state" in store_dict:
                unwrapped_model = self.accelerator.unwrap_model(self.model)
                unwrapped_model.load_state_dict(store_dict["model_state"])
                self.logging("Load model successfully")
            else:
                raise ValueError(
                    f"Model loading failed. self.model={self.model}, stored_dict_keys={store_dict.keys()}"
                )
        else:
            self.logging("Not load model")

        if load_optimizer:
            if self.optimizer and "optimizer_state" in store_dict:
                self.optimizer.load_state_dict(store_dict["optimizer_state"])
                self.logging("Load optimizer successfully")
            else:
                raise ValueError(
                    f"Model loading failed. self.model={self.optimizer}, stored_dict_keys={store_dict.keys()}"
                )
        else:
            self.logging("Not load optimizer")

        if load_history:
            history = store_dict.pop("history")
            self.reset_history(reset_all=True)
            if history is not None:
                self.history = history
            else:
                self.logging(
                    "Loaded history is None, reset history to empty.", level="WARNING"
                )

    def load_best_ckpt(
        self,
        load_optimizer: Optional[bool] = False,
        load_config: Optional[bool] = False,
        load_history: Optional[bool] = True,
    ):
        _task_dir = Path(self.config.task_dir)
        model_name = self.model.__class__.__name__
        ckpt_filepath = str(
            _task_dir.joinpath(
                CHECKPOINT_DIRNAME,
                BEST_CHECKPOINT_FILENAME_TEMPLATE.format(model_name),
            )
        )
        self.logging(f"Loading model from: {ckpt_filepath}")
        self.load(
            ckpt_filepath,
            load_model=True,
            load_optimizer=load_optimizer,
            load_config=load_config,
            load_history=load_history,
        )

    def save(self, path):
        self.accelerator.wait_for_everyone()
        self.logging(f"Dumping checkpoint into: {path}")
        store_dict = {}

        if self.model:
            model = self.accelerator.unwrap_model(self.model)
            model_state = model.state_dict()
            store_dict["model_state"] = model_state
        else:
            self.logging("No model state is dumped", logging.WARNING)

        if self.optimizer:
            store_dict["optimizer_state"] = self.optimizer.state_dict()
        else:
            self.logging("No optimizer state is dumped", logging.WARNING)

        store_dict["history"] = self.history

        torch.save(store_dict, path)

    def save_ckpt(self, identifier: Optional[str] = BEST_IDENTIFIER):
        ckpt_dir = os.path.join(self.config.task_dir, CHECKPOINT_DIRNAME)
        if not os.path.exists(ckpt_dir):
            os.makedirs(ckpt_dir)

        ckpt_name = CHECKPOINT_FILENAME_TEMPLATE.format(
            self.model.__class__.__name__, identifier
        )
        self.save(os.path.join(ckpt_dir, ckpt_name))

    def logging(self, msg: str, level: Optional[int] = logging.INFO):
        if self.config.only_master_logging:
            if self.accelerator.is_local_main_process:
                logger.log(level, msg)
        else:
            logger.log(level, msg)

    def is_master_node(self):
        return self.accelerator.is_local_main_process

    @classmethod
    def from_config(
        cls,
        config: OmegaConf,
        update_config: Optional[dict] = None,
        **kwargs,
    ):
        cls.logging(f"Initializing from configuration: {OmegaConf.to_yaml(config)}")

        # update
        if update_config is not None:
            for key, val in update_config.items():
                setattr(config, key, val)

        kwargs["initialize"] = kwargs.pop("initialize", True)
        kwargs["makedirs"] = kwargs.pop("makedirs", True)
        kwargs["dump_configfile"] = kwargs.pop("dump_configfile", True)
        return cls(config, **kwargs)

    @classmethod
    def from_configfile(
        cls,
        config_filepath: str,
        **kwargs,
    ):
        cls.logging(f"Initializing from configuration file: {config_filepath}")
        config = OmegaConf.load(config_filepath)

        return cls.from_config(config, **kwargs)

    @classmethod
    def from_taskdir(
        cls,
        task_dir: str,
        load_best_model: Optional[bool] = True,
        load_best_optimizer: Optional[bool] = False,
        load_config: Optional[bool] = False,
        **kwargs,
    ):
        _task_dir = Path(task_dir)
        config_filepath = str(_task_dir.joinpath(CONFIG_PARAMS_FILENAME).absolute())
        ins = cls.from_configfile(config_filepath, **kwargs)
        if load_best_model:
            ins.load_best_ckpt(
                load_optimizer=load_best_optimizer, load_config=load_config
            )
        return ins



from typing import Dict, Iterable, List, Optional, Union

from rex.utils.io import dump_json, load_json


class LabelEncoder(object):
    def __init__(self, initial_dict: Optional[Dict] = {}) -> None:
        self.label2id = initial_dict
        self.id2label = {idx: label for idx, label in enumerate(self.label2id)}

    def add(self, label: Union[str, int]):
        if label not in self.label2id:
            label_id = len(self.label2id)
            self.label2id[label] = label_id
            self.id2label[label_id] = label

    def update(self, labels):
        for label in labels:
            self.add(label)

    def encode(self, labels: List[Union[str, int]], update: Optional[bool] = False):
        if update:
            self.update(labels)

        if not all(label in self.label2id for label in labels):
            raise ValueError("Not all label are in this encoder")

        return list(map(lambda x: self.label2id[x], labels))

    def update_encode(self, labels):
        self.update(labels)
        return self.encode(labels)

    def update_encode_one(self, label):
        self.add(label)
        return self.encode([label])[0]

    def decode(self, ids: List[int]):
        if not all(idx in self.id2label for idx in ids):
            raise ValueError("Not all idx are in this encoder")
        return list(map(lambda x: self.id2label[x], ids))

    def convert_to_multi_hot(self, label_ids: Iterable[int]):
        labels = [0] * self.num_tags
        for label_id in label_ids:
            labels[label_id] = 1
        return labels

    def convert_to_one_hot(self, label_id: int):
        labels = [0] * self.num_tags
        labels[label_id] = 1
        return labels

    def __len__(self):
        return len(self.label2id)

    @property
    def num_tags(self):
        return len(self)

    @classmethod
    def from_pretrained(cls, filepath: str):
        label2id = load_json(filepath)
        return cls(initial_dict=label2id)

    def save_pretrained(self, filepath: str):
        dump_json(self.label2id, filepath)



import json
import pickle
from typing import Any, Iterable, List, Optional

from rex.utils.deprecation import deprecation_warning
from rex.utils.logging import logger


def dump_json(obj, filepath, **kwargs):
    with open(filepath, "wt", encoding="utf-8") as fout:
        json.dump(obj, fout, ensure_ascii=False, **kwargs)


def load_json(filepath, **kwargs):
    data = list()
    with open(filepath, "rt", encoding="utf-8") as fin:
        data = json.load(fin, **kwargs)
    return data


def dump_line_json(obj, filepath, **kwargs):
    deprecation_warning("dump_line_json", "dump_jsonlines")
    return dump_jsonlines(obj, filepath, **kwargs)


def dump_jsonlines(obj, filepath, **kwargs):
    with open(filepath, "wt", encoding="utf-8") as fout:
        for d in obj:
            line_d = json.dumps(d, ensure_ascii=False, **kwargs)
            fout.write("{}\n".format(line_d))


def load_line_json(filepath, **kwargs):
    deprecation_warning("load_line_json", "load_jsonlines")
    return load_jsonlines(filepath, **kwargs)


def load_jsonlines(filepath, **kwargs):
    data = list()
    with open(filepath, "rt", encoding="utf-8") as fin:
        for line in fin:
            line_data = json.loads(line.strip())
            data.append(line_data)
    return data


def dump_pickle(obj, filepath, **kwargs):
    with open(filepath, "wb") as fout:
        pickle.dump(obj, fout, **kwargs)


def load_pickle(filepath, **kwargs):
    data = None
    with open(filepath, "rb") as fin:
        data = pickle.load(fin, **kwargs)
    return data


def dump_csv(obj: Iterable[Any], filepath: str, delimiter: Optional[str] = "\t"):
    with open(filepath, "wt", encoding="utf-8") as fout:
        for d in obj:
            line_d = delimiter.join(d)
            fout.write("{}\n".format(line_d))


def load_csv(
    filepath: str,
    title_row: bool,
    title_keys: Optional[List[str]] = None,
    sep: Optional[str] = "\t",
) -> List:
    """load csv file

    Args:
        filepath: filepath to load
        title_row: has title in the first row or not?
                   If true, it'll return a list of dict where keys are from
                   the title, otherwise a list of str list.
        title_keys: if not `title_row`, you can set the title keys yourself.
        sep: separation char
    """
    data = list()
    title_keys = title_keys if title_keys else []
    with open(filepath, "rt", encoding="utf-8") as fin:
        for idx, line in enumerate(fin):
            line_data = line.strip().split(sep)
            if title_row and idx == 0:
                title_keys = line_data
                continue
            if title_keys:
                if len(title_keys) != len(line_data):
                    raise RuntimeError(
                        f"len of title keys: {title_keys}"
                        f" does not match the line data in line {idx + 1}"
                        f" in file: {filepath}"
                    )
                ins = {}
                for col, key in zip(line_data, title_keys):
                    ins[key] = col
            else:
                ins = line_data
            data.append(ins)
    return data


def load_embedding_file(filepath, encoding="utf-8", open_func=open, verbose=False):
    tokens = []
    token2vec = {}
    num_tokens = -1
    dim_emb = 0
    with open_func(filepath, "rt", encoding=encoding) as fin:
        for line_no, line in enumerate(fin):
            line = line.split()
            if line_no == 0:
                if len(line) == 2 and all(x.isdigit() for x in line):
                    num_tokens = int(line[0])
                    dim_emb = int(line[1])
                else:
                    dim_emb = len(line) - 1
                    tokens.append(line[0])
                    token2vec[line[0]] = list(map(float, line[1:]))
                continue
            # dimension checking
            if len(line) - 1 != dim_emb:
                continue
            tokens.append(line[0])
            token2vec[line[0]] = list(map(float, line[1:]))

    if num_tokens > 0 and num_tokens != len(tokens):
        logger.warning(
            f"emb file info num of tokens: {num_tokens}, while {len(tokens)} tokens are found"
        )

    if verbose:
        logger.info(f"Loading #Tokens: {len(tokens)}, Emb dim: {dim_emb}")

    return tokens, token2vec


def load_line_iterator(filepath):
    with open(filepath, "rt", encoding="utf-8") as fin:
        for line in fin:
            yield line


def load_line_json_iterator(filepath, **kwargs):
    deprecation_warning("load_line_json_iterator", "load_jsonlines_iterator")
    return load_jsonlines_iterator(filepath, **kwargs)


def load_jsonlines_iterator(filepath):
    for line in load_line_iterator(filepath):
        yield json.loads(line)


def dump_iterable(obj: Iterable, filepath: str):
    with open(filepath, "wt", encoding="utf-8") as fout:
        for line in obj:
            fout.write(f"{line}\n")



CONFIG_PARAMS_FILENAME = "task_params.yaml"
CHECKPOINT_DIRNAME = "ckpt"
CHECKPOINT_FILENAME_TEMPLATE = "{}.{}.pth"
BEST_IDENTIFIER = "best"
BEST_CHECKPOINT_FILENAME_TEMPLATE = "{}.best.pth"
LOG_FILENAME = "log.log"

from .relation_extraction import MCMLSentRelationClassificationTask
from .simple_task import SimpleTask



import os

import setuptools

from rex import __version__

readme_filepath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md")
with open(readme_filepath, "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="pytorch-rex",
    version=__version__,
    author="Tong Zhu",
    author_email="tzhu1997@outlook.com",
    description="A toolkit for Relation Extraction and more...",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/Spico197/REx",
    packages=setuptools.find_packages(exclude=["tests", "tests.*", "docs", "docs.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "numpy>=1.19.0",
        "scikit-learn>=0.21.3",
        "omegaconf>=2.0.6",
        "loguru>=0.5.3",
        "tqdm>=4.61.1",
    ],
    extras_require={
        "dev": ["pytest", "coverage", "black", "isort", "flake8", "pre-commit"]
    },
    include_package_data=True,
    entry_points={"console_scripts": ["rex = rex.cmds:main"]},
)



def construct_piecewise_mask(
    head_pos: int, tail_pos: int, seq_len: int, max_seq_len: int
):
    assert 0 <= head_pos < seq_len
    assert 0 <= tail_pos < seq_len

    mask = [0] * max_seq_len
    for i in range(0, max_seq_len):
        if 0 <= i < min(head_pos, tail_pos):
            mask[i] = 1
        elif min(head_pos, tail_pos) <= i < max(head_pos, tail_pos):
            mask[i] = 2
        elif max(head_pos, tail_pos) <= i < min(max_seq_len, seq_len):
            mask[i] = 3
        else:
            mask[i] = 0
    return mask



from typing import Callable, Iterable, List, Optional, Tuple

import numpy as np

from rex.utils.io import dump_iterable, load_embedding_file


def _convert_list_str_to_list(string: str, item_type: Callable):
    ret = []
    if string.startswith("["):
        string = string[1:]
    if string.endswith("]"):
        string = string[:-1]

    for item in string.split(","):
        ret.append(item_type(item.strip()))

    return ret


def get_pad_mask(token_len, max_len, token_mask=1, pad_mask=0):
    if token_len > max_len:
        token_len = max_len
    mask = [token_mask] * token_len + [pad_mask] * (max_len - token_len)
    return mask


def get_pad_token(tokens, max_len, pad_token):
    tokens_len = len(tokens)
    tokens = tokens[:max_len]
    tokens = tokens + [pad_token] * (max_len - tokens_len)
    return tokens


class Vocab(object):
    def __init__(
        self,
        pad: Optional[str] = "[PAD]",
        unk: Optional[str] = "[UNK]",
        include_pad: Optional[bool] = True,
        include_unk: Optional[bool] = True,
        embedding_dim: Optional[int] = 300,
        init_pad_unk_emb: Optional[bool] = False,
    ) -> None:
        self.pad = pad
        self.unk = unk
        self.pad_idx = None
        self.unk_idx = None

        self.token2id = {}
        self.id2token = {}
        self.weights = []

        if include_pad:
            token_idx = len(self.token2id)
            self.token2id[pad] = token_idx
            self.id2token[token_idx] = pad
            self.pad_idx = self.token2id[pad]
            if init_pad_unk_emb:
                self.weights.append(np.random.randn(embedding_dim).tolist())

        if include_unk:
            token_idx = len(self.token2id)
            self.token2id[unk] = token_idx
            self.id2token[token_idx] = unk
            self.unk_idx = self.token2id[unk]
            if init_pad_unk_emb:
                self.weights.append(np.random.randn(embedding_dim).tolist())

    def add(self, token: str, weights: Optional[List[float]] = None):
        if token not in self.token2id:
            token_id = len(self.token2id)
            self.token2id[token] = token_id
            self.id2token[token_id] = token

            if token == self.pad:
                self.pad_idx = token_id
            elif token == self.unk:
                self.unk_idx = token_id

            if weights is not None:
                if token_id != len(self.weights):
                    raise ValueError(
                        "Vocab does not match weights, check `init_pad_unk_emb` option"
                    )
                self.weights.append(weights)

    def update(self, tokens: List[str]):
        for token in tokens:
            self.add(token)

    def convert_tokens_to_ids(self, tokens: List[str]):
        return list(map(lambda x: self.token2id.get(x, self.unk_idx), tokens))

    def convert_ids_to_tokens(self, ids: List[int]):
        if not all(idx in self.id2token for idx in ids):
            raise ValueError("Not all idx are in this vocab")
        return list(map(lambda x: self.id2token[x], ids))

    def update_convert_tokens_to_ids(self, tokens: List[str]):
        for token in tokens:
            self.add(token)
        return self.convert_tokens_to_ids(tokens)

    def encode(
        self, tokens: Iterable, max_seq_len: int, update: Optional[bool] = False
    ) -> Tuple[List[int]]:
        """convert tokens into ids by padding or cutting

        Args:
            update: whether to add tokens into vocab

        Returns:
            token_ids: token ids after encoding
            mask: padding mask
        """
        mask = get_pad_mask(len(tokens), max_seq_len, 1, 0)
        tokens = get_pad_token(tokens, max_seq_len, self.pad)
        if update:
            return self.update_convert_tokens_to_ids(tokens), mask
        else:
            return self.convert_tokens_to_ids(tokens), mask

    def clear_all(self):
        self.token2id.clear()
        self.id2token.clear()

    def __len__(self):
        return len(self.token2id)

    @property
    def size(self):
        return len(self)

    @classmethod
    def from_pretrained(
        cls, filepath, include_weights: Optional[bool] = False, **kwargs
    ):
        v = cls(**kwargs)
        tokens, token2vec = load_embedding_file(filepath)
        for token in tokens:
            vec = None
            if include_weights:
                vec = token2vec[token]
            v.add(token, weights=vec)
        return v

    def save_pretrained(self, filepath, dump_weights: Optional[bool] = False):
        vocabs = []
        for token_id in range(self.size):
            if dump_weights:
                weight_str = " ".join(map(str, self.weights[token_id]))
                vocabs.append(f"{self.id2token[token_id]} {weight_str}")
            else:
                vocabs.append(self.id2token[token_id])
        dump_iterable(vocabs, filepath)
